/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/include/HSystem.hh,v 1.40 2002/04/24 18:05:08 okumoto Exp $
 */

// -*- C++ -*-  Sun Nov 29 1998
// Eugene Skepner <eu@qub.com>
// $Id: HSystem.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef HSystem_hh
#define HSystem_hh

#include "LanguageInterface.hh"
#include "String.hh"

#include <iostream>
#include <list>
#include <set>

/**
 * Represent a libh subsystem object interface generator.
 */
class HSystem {
    /*
     * Types used to simplify code.
     */
    typedef LanguageInterface::Object::ClassDescription		ClassDesc;
    typedef LanguageInterface::Object::MethodDescription	MethodDesc;
    typedef LanguageInterface::Object::CallbackDescription	CallBackDesc;

    typedef list<const ClassDesc *>	Descriptions;
    typedef set<string>			Includes;

    Descriptions		descriptions;
    Includes			includes;

    /**
     * Output header and includes. The output text contains a
     * warning that it is a generated source file.
     *
     * @param out	stream to send text.
     */
    void gen_header(ostream &out) const;

    /**
     * Output a class definition for each class description.
     *
     * @param out	stream to send text.
     */
    void gen_classes(ostream &out) const;

    /**
     * Output a function to register each class description with the TCL
     * interpreture.
     *
     * @param out	stream to send text.
     */
    void gen_register_command_access(ostream &out) const;

    /**
     * Output comment marking end of file.
     *
     * @param out	stream to send text.
     */
    void gen_footer(ostream &out) const;

    /**
     * Output a class definition for the given cd.
     *
     * @param out	stream to send text.
     * @param cd	Class description.
     */
    void gen_class(ostream &out, const ClassDesc &cd) const;

    void gen_methodCommand(ostream &out) const;

    void gen_dynamic_methods(ostream &out, const ClassDesc &cd) const;

    void gen_callback_interfaces(ostream &out, const ClassDesc &cd) const;

    inline string
    paramlist(const MethodDesc &method, int max) const
    {
        string	result;
        result += method.name_to_call;
        result += "( ";
	for (int i = 0; i < max; ++i) {
	    result += 'a';
	    result += to_string(i);
	    if (i < max - 1) {
		result += ", ";
	    }
	}
        result += " )";
	return result;
    }

    inline bool
    arg_islist(const MethodDesc &method, int pos) const
    {
	return (method.args[pos] == MethodDesc::vtList ||
		method.args[pos] == MethodDesc::vtListRef ||
		method.args[pos] == MethodDesc::vtListPtr);
    }

    /**
     */
    inline string
    arg_type(const MethodDesc &method, int pos) const
    {
	return
	    LanguageInterface::method_argument_type(method.args[pos], false) +
	    " a" +
	    to_string(pos);
    }

    /**
     */
    inline string
    result_type(const MethodDesc &method) const
    {
	return LanguageInterface::method_result_type(method.result, false);
    }

    /**
     * Parse *.cd.cc file for ClassDescriptions.
     *
     * @param filename	Name of input file.
     *
     * @note This func has yet to be written.
     */
    void parse_ClassDesc(const string &filename);

    /**
     */
    string gen_result_to_obj(MethodDesc::Value t) const;

    /**
     */
    string gen_method_argument(MethodDesc::Value t, int argno, int method_argno, bool last_arg_in_set_method) const;

    void gen_method(ostream& out, const ClassDesc &cd, const MethodDesc &method) const;

  public:
    /**
     * Name of subsystem.
     *
     * @note This should be made private But this would require modifications
     * outside of HSystem.hh.
     */
    const string	name;

    /**
     * Constructor.
     *
     * @param aName	Name of subsystem.
     */
    HSystem(const string &aName)
	: name(aName) {}

    /**
     * Copy constructor.  XXX This is broken!  None of the other state is
     * copied!
     *
     * @param aHSystem	HSystem object to copy.
     */
    HSystem(const HSystem &aHSystem)
	: name(aHSystem.name) {}

    /**
     * LessThan operator.
     *
     * @param rhs	Right hand side of the less than operator.
     */
    bool operator<(const HSystem &rhs) const { return name < rhs.name; }

    /**
     * Equivalence operator.
     *
     * @param rhs	Right hand side of the equivalence operator.
     */
    bool operator==(const HSystem &rhs) const { return name == rhs.name; }

    /**
     * Adds the class description and include file to the system definition.
     *
     * @param cd
     * @param include_file
     *
     * @see find_classes_descriptions.pl
     */
    void
    add_class(const ClassDesc *cd, const string &include_file)
    {
	descriptions.push_back(cd);
	includes.insert(include_file);
    }

    /**
     * Generate libh subsystem TCL interface text.
     *
     * @param out	stream to send text.
     */
    void gen_file(ostream &out) const;
};

/**
 * a set of systems
 */
typedef set<HSystem> Systems;

/**
 * Initialize a subsystem and register it.
 *
 * The implementation of this function must add the subsystem to the
 * Systems set. The subsystem added must include all classes
 * description the subsystems wishes to make accessible from TCL.
 *
 * @note this function is created by find_classes_descriptions.pl for
 * each subsystem.
 *
 * @see find_classes_descriptions.pl
 * @param systems	the set of HSystem objects to add the subsystem to
 */
void build_systems(Systems &systems);

#endif // HSystem_hh
