#ifndef Dim_hh
#define Dim_hh
/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/include/hui/Dim.hh,v 1.1 2002/05/13 16:29:41 okumoto Exp $
 */
#include "Container.hh"
#include "Pointer.hh"

#include <limits>
#include <vector>

typedef short			Dimension;
typedef vector<Dimension>	DimVector;
typedef Container<DimVector>	DimVector_C;

/**
 * Represents a pair of Dimensions used together.
 */
class DimensionPair {
protected:
    Dimension	first;
    Dimension	second;

public:
    /**
     * Constructor.
     */
    DimensionPair(Dimension a, Dimension b)
	: first(a), second(b)
    {
    }

    /**
     * Constructor.
     */
    DimensionPair(const DimVector &v)
    {
	switch (v.size()) {
	case 0: first = 0; second = 0; break;
	case 2: first = v[0]; second = v[1]; break;
	default:
	    assert(0);	/* We lost data */
	}
    }

    /**
     * Destructor.
     */
    virtual ~DimensionPair() {}

    /**
     *
     */
    operator Pointer<DimVector_C> () const
    {
	Pointer<DimVector_C> r = new DimVector_C(2);
	(*r)[0] = first;
	(*r)[1] = second;
	return r;
    }
};

class Position : public DimensionPair {
public:
    /**
     * Constructor.
     */
    Position(Dimension aX = 0, Dimension aY = 0)
	: DimensionPair(aX, aY)
    {
    }

    /**
     * Constructor.
     */
    Position(const DimVector &v)
	: DimensionPair(v)
    {
    }

    Dimension
    x() const
    {
	return first;
    }

    void
    x(Dimension aX)
    {
	first = aX;
    }

    Dimension
    y() const
    {
	return second;
    }

    void
    y(Dimension aY)
    {
	second = aY;
    }

    Position
    operator+(const Position &rhs) const
    {
	return(Position(x() + rhs.x(), y() + rhs.y()));
    }
};

class Size : public DimensionPair {
public:
    /**
     * Constructor.
     */
    Size(const Dimension aWidth = 0, const Dimension aHeight = 0)
	: DimensionPair(aWidth, aHeight)
    {
    }

    /**
     * Constructor.
     */
    Size(const DimVector &v)
	: DimensionPair(v)
    {
    }

    Dimension
    width() const
    {
	return first;
    }

    void
    width(Dimension aWidth)
    {
	first = aWidth;
    }

    Dimension
    height() const
    {
	return second;
    }

    void
    height(Dimension aHeight)
    {
	second = aHeight;
    }

    static Dimension
    maxWidth()
    {
	return numeric_limits<Dimension>::max();
    }

    static Dimension
    maxHeight()
    {
	return numeric_limits<Dimension>::max();
    }

    Size &
    operator+=(const Size &rhs)
    {
	if (rhs.width() != maxWidth() && width() != maxWidth())
	    width(width() + rhs.width());
	else
	    width(maxWidth());

	if (rhs.height() != maxHeight() && height() != maxHeight())
	    height(height() + rhs.height());
	else
	    height(maxHeight());
	return(*this);
    }

    Size &
    operator+=(const Position &rhs)
    {
	if (width() != maxWidth())
	    width(width() + rhs.x());
	if (height() != maxHeight())
	    height(height() + rhs.y());
	return(*this);
    }

    Size
    operator+(const Size &rhs) const
    {
	Size result = *this;
	result += rhs;
	return(result);
    }

    Size
    operator+(const Position &rhs) const
    {
	Size result = *this;
	result += rhs;
	return(result);
    }

    Size &
    operator-=(const Size &rhs)
    {
	if (width() != maxWidth())
	    width(width() - rhs.width());
	if (height() != maxHeight())
	    height(height() - rhs.height());
	return(*this);
    }

    Size &
    operator-=(const Position &rhs)
    {
	if (width() != maxWidth())
	    width(width() - rhs.x());
	if (height() != maxHeight())
	    height(height() - rhs.y());
	return(*this);
    }

    Size
    operator-(const Size &rhs) const
    {
	Size result = *this;
	result -= rhs;
	return(result);
    }

    Size
    operator-(const Position &rhs) const
    {
	Size result = *this;
	result -= rhs;
	return(result);
    }
};

#endif /* Dim_hh */
