/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/include/hui/Hui.hh,v 1.39 2002/05/13 16:29:41 okumoto Exp $
 */

// -*- C++ -*-  Tue Dec  1 1998
// Eugene Skepner <eu@qub.com>
// $Id: Hui.hh,v 1.1.1.1 2000/02/15 16:30:35 alex Exp $

#ifndef Hui_hh
#define Hui_hh

#include "LanguageInterface.hh"
#include "Container.hh"
#include "Log.hh"
#include "hui/Dim.hh"

#include <stdexcept>
#include <vector>

#define ARRANGE_FREE		0
#define ARRANGE_HORIZONTAL	1
#define ARRANGE_VERTICAL	2

/**
 * Represent a libh user interface (abstract base class).
 */
class Hui : virtual public LanguageInterface::Object {
    CLASS_METHODS_DESCRIPTIONS_DECLARATION;

protected:
    class Error : public runtime_error {
    public:
	    Error(const string &what_arg) : runtime_error(what_arg) {}
    };

    /*
     * Forward declarations
     */
    class ButtonWidget;
    class CheckBoxesWidget;
    class ContainerWidget;
    class DialogWidget;
    class EmbeddedWidget;
    class GroupWidget;
    class InputLineWidget;
    class LabelWidget;
    class ListBoxWidget;
    class MenuItemWidget;
    class MenuSubmenuWidget;
    class ProgressIndicatorWidget;
    class RadioMenuWidget;
    class TextEditWidget;
    class Widget;
    class WindowWidget;

    /** Represents the types of interfaces. */
    enum Type {tAuto, tText, tGraphics};

    typedef LanguageInterface::CallbackData	CallBackData;

    typedef Pointer<ButtonWidget>		ButtonWidget_P;
    typedef Pointer<CallBackData>		CallBackData_P;
    typedef Pointer<CheckBoxesWidget>		CheckBoxesWidget_P;
    typedef Pointer<ContainerWidget>		ContainerWidget_P;
    typedef Pointer<DialogWidget>		DialogWidget_P;
    typedef Pointer<EmbeddedWidget>		EmbeddedWidget_P;
    typedef Pointer<GroupWidget>		GroupWidget_P;
    typedef Pointer<Hui>			Hui_P;
    typedef Pointer<InputLineWidget>		InputLineWidget_P;
    typedef Pointer<LabelWidget>		LabelWidget_P;
    typedef Pointer<ListBoxWidget>		ListBoxWidget_P;
    typedef Pointer<MenuItemWidget>		MenuItemWidget_P;
    typedef Pointer<MenuSubmenuWidget>		MenuSubmenuWidget_P;
    typedef Pointer<ProgressIndicatorWidget>	ProgressIndicatorWidget_P;
    typedef Pointer<RadioMenuWidget>		RadioMenuWidget_P;
    typedef Pointer<TextEditWidget>		TextEditWidget_P;
    typedef Pointer<Widget>			Widget_P;
    typedef Pointer<WindowWidget>		WindowWidget_P;

    typedef list<MenuItemWidget_P>		Items;

    class WidgetCallbackData : public LanguageInterface::SpecificCallbackData {
	Hui	&mHui;
	Widget	&mWidget;
	string	mResult;

    public:
	/**
	 * Constructor.
	 */
	WidgetCallbackData(Hui &aHui,
			   Widget &aWidget,
			   CallBackData_P aCallbackData)
	    : LanguageInterface::SpecificCallbackData(aCallbackData),
	      mHui(aHui),
	      mWidget(aWidget)
	{
	}

	/**
	 * @note This function is required because we derive from
	 * LanguageInterface::SpecificCallbackData.
	 */
	const char *
	typeid_name() const
	{
	    return(typeid(Hui).name());
	}

	const string &
	activate(bool call_callback_completed)
	{
	    try {
		if (callbackData()) {
		    callbackData()->activate(*this);
		    if (call_callback_completed)
			widget()->callback_completed(*this);
		}
	    } catch (exception &err) {
		LOG(ERROR, "Hui::WidgetCallbackData::activate: " << err.what());
	    }
	    return(mResult);
	}

	template <typename Arg> void
	activate_with_arg(Arg arg)
	{
	    try {
		if (callbackData())
		    callbackData()->activate(*this, arg);
	    } catch (exception &err) {
		LOG(ERROR, "Hui::WidgetCallbackData::activate_with_arg: " << err.what());
	    }
	}

	string &
	result()
	{
	    return mResult;
	}

	const string &
	result() const
	{
	    return mResult;
	}

	Hui_P
	hui()
	{
	    return &mHui;
	}

	Widget_P
	widget()
	{
	    return &mWidget;
	}
    };

    class Widget : virtual public LanguageInterface::Object {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	string			mName;
	WidgetCallbackData	mOnEnter;
	WidgetCallbackData	mOnLeave;

    protected:
	/**
	 * Constructor.
	 */
	Widget(Hui &aHui, const string &aName)
	    : mName(aName),
	      mOnEnter(aHui, *this, 0),
	      mOnLeave(aHui, *this, 0)
	{
	}

    public:
	/**
	 * Destructor.
	 */
	virtual ~Widget() {}

	virtual string
	name(const string &aName)
	{
	    if (aName.length() == 1 && aName[0] == '\0') {
		return(mName);
	    } else {
		string oldName = mName;
		mName = aName;
		return(oldName);
	    }
	}

	virtual const string &
	name() const
	{
	    return mName;
	}

	virtual Widget_P
	byName(const string &aName) const
	{
	    return(name() == aName ?
		   const_cast<Widget*>(this)
		   : 0);
	}

	virtual void show() = 0;
	virtual void hide() = 0;
	virtual bool hidden() const = 0;

	virtual void disable() = 0;
	virtual void enable() = 0;
	virtual bool disabled() const = 0;

	virtual void
	enter()
	{
	    mOnEnter.activate(false);
	}

	virtual void
	leave()
	{
	    mOnLeave.activate(false);
	}

	virtual void
	onEnter(CallBackData_P aCallbackData)
	{
	    mOnEnter.callbackData(aCallbackData);
	}

	virtual CallBackData_P
	onEnter() const
	{
	    return(mOnEnter.callbackData());
	}

	virtual void
	onLeave(CallBackData_P aCallbackData)
	{
	    mOnLeave.callbackData(aCallbackData);
	}

	virtual CallBackData_P
	onLeave() const
	{
	    return(mOnLeave.callbackData());
	}

	virtual void move(const Position &aPosition) = 0;
	virtual Position position() const = 0;
	virtual void resize(const Size &aSize) = 0;
	virtual Size size() const = 0;

	virtual void
	callback_completed(const WidgetCallbackData &/*aCallbackData*/)
	{
		/* XXX empty? what is this for? */
	}
    };

    class EmbeddedWidget : virtual public Widget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	EmbeddedWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName)
	{
	}

    };

    class GroupWidget : virtual public Widget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	GroupWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName)
	{
	}

    public:
	virtual LabelWidget_P
	label(const string &aName,
	      const string &aText,
	      const Position &aPosition) = 0;

	virtual ButtonWidget_P
	button(const string &aName,
	       const string &aText,
	       CallBackData_P aCallbackData,
	       const string &aButtonType,
	       const Position &aPosition,
	       const string &aComment) = 0;

	virtual InputLineWidget_P
	inputLine(const string &aName,
		  const string &aInitialText,
		  int aMaxLength,
		  const string &aValidator,
		  const string &aTitle,
		  const Position &aPosition,
		  const string &aComment) = 0;

	virtual CheckBoxesWidget_P
	checkBoxes(const string &aName,
		  const vector<string> &aTexts,
		  const vector<bool> &aInitialValues,
		  const string &aTitle,
		  const Position &aPosition,
		  const vector<string> &aComments) = 0;

	virtual RadioMenuWidget_P
	radioMenu(const string &aName,
		const vector<string> &aTexts,
		const int aInitialValue,
		const string &aTitle,
		const Position &aPosition,
		const vector<string> &aComments) = 0;

	virtual ListBoxWidget_P
	listBox(const string &aName,
		const vector<string> &aTexts,
		const int aInitiallySelected,
		const string &aTitle,
		const Position &aPosition,
		const string &aComment) = 0;

	virtual TextEditWidget_P
	textEdit(const string &aName,
	      const string &aInitialText,
	      bool aAllowEdit,
	      const string &aTitle,
	      const Position &aPosition,
	      const string &aComment) = 0;

	virtual ProgressIndicatorWidget_P
	progressIndicator(const string &aName,
		float aMin,
		float aMax,
		float aInitial,
		const string &aTitle,
		const Position &aPosition,
		const string &aComment) = 0;

	virtual ContainerWidget_P
	container(const string &aName,
		const Position &aPosition,
		const string &aAlignment,
		const string &aComment) = 0;

	virtual void arrange() = 0;
	virtual void arrange(const string &aAlignmen, bool arrange_new) = 0;

	virtual Widget_P byName(const string &aName) const = 0;

	virtual Widget_P current() const = 0;
	virtual void current(EmbeddedWidget_P aWidget) = 0;

	virtual void erase() = 0;
    };

    class ContainerWidget : virtual public GroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	ContainerWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      GroupWidget(aHui, aName)
	{
	}
    };

    class ToplevelWidget : virtual public GroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	ToplevelWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      GroupWidget(aHui, aName)
	{
	}

    public:
	virtual void title(const string &aTitle) = 0;
	virtual string title() const = 0;
    };

    class LabelWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	LabelWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName)
	{
	}

    public:
	virtual void text(const string &aText) = 0;
	virtual string text() const = 0;
    };

    class ButtonWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	WidgetCallbackData mOnClick;

    protected:
	/**
	 * Constructor.
	 */
	ButtonWidget(Hui &aHui,
		     const string &aName,
		     CallBackData_P aCallbackData)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      mOnClick(aHui, *this, aCallbackData)
	{
	}

    public:
	virtual void text(const string &aText) = 0;
	virtual string text() const = 0;

	virtual void
	click()
	{
	    mOnClick.activate(true);
	}

	virtual void
	onClick(CallBackData_P aCallbackData)
	{
	    mOnClick.callbackData(aCallbackData);
	}

	virtual CallBackData_P
	onClick() const
	{
	    return(mOnClick.callbackData());
	}
    };

    class InputLineWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	int mMaxLength;

    protected:
	/**
	 * Constructor.
	 */
	InputLineWidget(Hui &aHui, const string &aName, int aMaxLength)
		: Widget(aHui, aName),
		  EmbeddedWidget(aHui, aName),
		  mMaxLength(aMaxLength)
	{
	}

	int
	maxLength() const
	{
	    return(mMaxLength);
	}

    public:
	virtual void title(const string &aTitle) = 0;
	virtual string title() const = 0;
	virtual void text(const string &aText) = 0;
	virtual string text() const = 0;
    };

    class ButtonsGroupWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	WidgetCallbackData mOnClick;

    protected:
	/**
	 * Constructor.
	 */
	ButtonsGroupWidget(Hui &aHui, const string &aName)
		: Widget(aHui, aName),
		  EmbeddedWidget(aHui, aName),
		  mOnClick(aHui, *this, 0)
	{
	}

    public:
	virtual void text(int boxno, const string &aText) = 0;
	virtual string text(int boxno) const = 0;
	virtual void title(const string &aTitle) = 0;
	virtual string title() const = 0;
	virtual void disable_item(int boxno) = 0;
	virtual void enable_item(int boxno) = 0;
	virtual bool item_disabled(int boxno) const = 0;
	virtual int numberOfItems() const = 0;

	virtual void
	click(int boxno)
	{
	    mOnClick.activate_with_arg(boxno);
	}

	virtual void
	onClick(CallBackData_P aCallbackData)
	{
	    mOnClick.callbackData(aCallbackData);
	}

	virtual CallBackData_P
	onClick() const
	{
	    return(mOnClick.callbackData());
	}
    };

    class CheckBoxesWidget : virtual public ButtonsGroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	CheckBoxesWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      ButtonsGroupWidget(aHui, aName)
	{
	}

    public:
	virtual bool state(int boxno, int aNewState) = 0;
	virtual bool state(int boxno) const = 0;
	virtual bool toggle(int boxno) = 0;
    };

    class RadioMenuWidget : virtual public ButtonsGroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	RadioMenuWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      ButtonsGroupWidget(aHui, aName)
	{
	}

    public:
	virtual void set(int itemno) = 0;
	virtual int set() const = 0;
    };

    class ListBoxWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	WidgetCallbackData mOnSelect;

    protected:
	/**
	 * Constructor.
	 */
	ListBoxWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      mOnSelect(aHui, *this, 0)
	{
	}

    public:
	virtual void title(const string &aTitle) = 0;
	virtual string title() const = 0;
	virtual int numberOfItems() const = 0;
	virtual string text(int aItemNo) const = 0;
	virtual void text(int aItemNo, const string &aText) = 0;
	virtual void select(int aItemNo) = 0;
	virtual int select() const = 0;
	virtual void add(int aBefore, const string &aText) = 0;
	virtual void remove(int aItemNo) = 0;
	virtual void removeAll() = 0;

	virtual void
	highlighted(int itemno)
	{
	    mOnSelect.activate_with_arg(itemno);
	}

	virtual void
	onSelect(CallBackData_P aCallbackData)
	{
	    mOnSelect.callbackData(aCallbackData);
	}

	virtual CallBackData_P
	onSelect() const
	{
	    return(mOnSelect.callbackData());
	}
    };

    class TextEditWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	TextEditWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName)
	{
	}

    public:
	virtual void title(const string &aTitle) = 0;
	virtual string title() const = 0;
	virtual void text(const string &aText) = 0;
	virtual string text() const = 0;
    };

    class ProgressIndicatorWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	ProgressIndicatorWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName)
	{
	}

    public:
	virtual void title(const string &aTitle) = 0;
	virtual string title() const = 0;
	virtual void set(float aValue) = 0;
	virtual float set() const = 0;
    };

    class MenuItemWidget : virtual public EmbeddedWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	WidgetCallbackData mOnClick;

    protected:
	/**
	 * Constructor.
	 */
	MenuItemWidget(Hui &aHui,
		       const string &aName,
		       CallBackData_P aCallbackData)
		: Widget(aHui, aName),
		  EmbeddedWidget(aHui, aName),
		  mOnClick(aHui, *this, aCallbackData) {
	}

    public:
	virtual void text(const string &aText) = 0;
	virtual string text() const = 0;
	virtual string key() const = 0;

	virtual void
	click()
	{
	    mOnClick.activate(true);
	}

	virtual void
	onClick(CallBackData_P aCallbackData)
	{
	    mOnClick.callbackData(aCallbackData);
	}

	virtual CallBackData_P
	onClick() const
	{
	    return(mOnClick.callbackData());
	}

	virtual void
	move(const Position &/*aPosition*/)
	{
	}

	virtual Position
	position() const
	{
	    return Position();
	}

	virtual void
	resize(const Size &/*aSize*/)
	{
	}

	virtual Size
	size() const
	{
	    return Size();
	}

	virtual bool
	removeItem(MenuItemWidget_P /*aItem*/)
	{
	    return(false);
	}
    };

    class MenuSeparatorWidget : virtual public MenuItemWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	MenuSeparatorWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      MenuItemWidget(aHui, aName, 0)
	{
	}

    public:
	virtual void
	text(const string &/*aText*/)
	{
	}

	virtual string
	text() const
	{
	    return string();
	}

	virtual string
	key() const
	{
	    return string();
	}

	virtual void
	click()
	{
	}
    };

    class MenuButtonWidget : virtual public MenuItemWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	string mText;
	string mKey;
	string mComment;

    protected:
	/**
	 * Constructor.
	 */
	MenuButtonWidget(Hui &aHui,
			 const string &aName,
			 const string &aText,
			 CallBackData_P aCallbackData,
			 const string &aKey,
			 const string &aComment)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      MenuItemWidget(aHui, aName, aCallbackData),
	      mText(aText),
	      mKey(aKey),
	      mComment(aComment)
	{
	}

	virtual void
	keyText(const string &aKey)
	{
	    mKey = aKey;
	}

	virtual void
	setText(const string &aText)
	{
	    mText = aText;
	}

    public:
	virtual void key(const string &aKey) = 0;

	virtual string
	key() const
	{
	    return(mKey);
	}

	virtual string
	text() const
	{
	    return(mText);
	}
    };

    /**
     */
    class WindowWidget : virtual public ToplevelWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    protected:
	/**
	 * Constructor.
	 */
	WindowWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      GroupWidget(aHui, aName),
	      ToplevelWidget(aHui, aName)
	{
	}
    };

    /**
     */
    class DialogWidget : virtual public ToplevelWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	virtual const string &result() const = 0;
	virtual const string &exec() = 0;

    protected:
	/**
	 * Constructor.
	 */
	DialogWidget(Hui &aHui, const string &aName)
	    : Widget(aHui, aName),
	      GroupWidget(aHui, aName),
	      ToplevelWidget(aHui, aName)
	{
	}
    };

    /**
     */
    class MenuSubmenuWidget : virtual public MenuItemWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	string	mText;
	Items	mItems;
	string	mComment;

    protected:
	/**
	 * Constructor.
	 */
	MenuSubmenuWidget(Hui &aHui,
			  const string &aName,
			  const string &aText,
			  const string &aComment)
	    : Widget(aHui, aName),
	      EmbeddedWidget(aHui, aName),
	      MenuItemWidget(aHui, aName, 0),
	      mText(aText),
	      mComment(aComment)
	{
	}

	virtual void
	addItem(MenuItemWidget_P aBefore, MenuItemWidget_P aItem)
	{
	    mItems.insert(find(mItems.begin(), mItems.end(), aBefore), aItem);
	}


	virtual bool
	removeItem(MenuItemWidget_P aItem)
	{
	    bool result = false;
	    Items::iterator i = find(mItems.begin(), mItems.end(), aItem);
	    if (i != mItems.end()) {
		mItems.erase(i);
		result = true;
	    } else {
		for (Items::iterator i = mItems.begin(); i != mItems.end() && !result; ++i) {
		    result = (*i)->removeItem(aItem);
		}
	    }
	    return(result);
	}

	virtual void
	setText(const string &aText)
	{
	    mText = aText;
	}

    public:
	virtual MenuItemWidget_P
	addButton(const string &aName,
		 const string &aText,
		 CallBackData_P aCallbackData,
		 const string &aKey,
		 const string &aComment,
		 MenuItemWidget_P aBefore) = 0;

	virtual MenuItemWidget_P
	addSeparator(const string &aName, MenuItemWidget_P aBefore) = 0;

	virtual MenuItemWidget_P
	addSubmenu(const string &aName,
		const string &aText,
		const string &aComment,
		MenuItemWidget_P aBefore) = 0;

	virtual void
	add(MenuItemWidget_P aBefore, MenuItemWidget_P aItem) = 0;

	virtual void
	remove(MenuItemWidget_P aItem) = 0;

	virtual Widget_P
	byName(const string &aName) const
	{
	    Widget_P result = 0;
	    for (Items::const_iterator i = mItems.begin(); i != mItems.end() && !result; ++i) {
		result = (*i)->byName(aName);
	    }
	    return(result);
	}

	virtual string
	text() const
	{
	    return(mText);
	}

	virtual string
	key() const
	{
	    return string();
	}

	virtual const Items &
	items() const
	{
	    return(mItems);
	}

	virtual Items &
	items()
	{
	    return(mItems);
	}
    };

public:
    /**
     * Constructor.
     */
    Hui() {}

    /**
     * Create an instance of an UI object given a type.
     *
     * @param aType		type of interface to build.
     */
    static Hui_P
    open(const string &aType);

    /**
     * Create an instance of an UI object given a type.
     *
     * @param type		type of interface to build.
     */
    static Hui_P
    open(Type type = tAuto);

    /**
     * Run the application.
     */
    virtual void run() = 0;

    /**
     * Create a widget by name.
     */
    virtual Widget_P
    byName(const string &aName) = 0;

    /**
     * Create a window.
     */
    virtual WindowWidget_P
    window(const string &aName,
	   const string &aTitle,
	   const Position &aPosition,
	   const Size &aSize,
	   const string &aAlignment,
	   bool no_menu = false) = 0;

    /**
     * Create a dialog.
     */
    virtual DialogWidget_P
    dialog(Widget_P aParent,
	   const string &aName,
	   const string &aTitle,
	   const Position &aPosition,
	   const Size &aSize,
	   const string &aAlignment) = 0;

    /**
     * Create a menu bar.
     */
    virtual MenuSubmenuWidget_P
    menuBar() = 0;

    /**
     */
    virtual void processPendingEvents() = 0;

    /**
     * Returns an empty string if user pressed cancel button.
     */
    virtual string
    fileSelectionDialog(const string &aTitle, const string &aWildCard) = 0;

    /**
     * Returns type of UI.
     */
    virtual const char *type() const = 0;

    /**
     * Returns size of UI.
     */
    virtual Size size() const = 0;

    /**
     * Exit the application.
     */
    virtual void exit(int status) = 0;

    /**
     * Destructor.
     */
    virtual ~Hui() {}
};

#endif /* Hui_hh */
