/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/include/hui/HuiTVision.hh,v 1.54 2002/05/13 16:29:41 okumoto Exp $
 */

// -*- C++ -*-  Wed Dec  2 1998
// Eugene Skepner <eu@qub.com>
// $Id: HuiTVision.hh,v 1.1.1.1 2000/02/15 16:30:35 alex Exp $

#ifndef HuiTVision_hh
#define HuiTVision_hh

#include "Hui.hh"

#include <map>

//======================================================================

#define Uses_TApplication
#if defined(H_BUILD_SYSTEMS) || defined(H_LIB_TCL_INTERFACE)
#define cstrlen tvision_cstrlen // hack to allow both TVision and Qt headers to be included together into the same module
#endif

#define Uses_TDialog
#define Uses_TLabel
#define Uses_TSItem

#define Uses_TCluster
#define Uses_TMenuBar
#define Uses_TMenuItem
#define Uses_TSubMenu

#include <tvision/tv.h>
#if defined(H_BUILD_SYSTEMS) || defined(H_LIB_TCL_INTERFACE)
#undef cstrlen
#endif

#define	TRUE	1
#define	FALSE	0

/**
 * Represent a libh user interface (using the TVision library)
 */
class HuiTVision : public TApplication, public Hui {
    CLASS_METHODS_DESCRIPTIONS_DECLARATION;

    class HuiTButton;
    class HuiTCheckBoxes;
    class HuiTDialog;
    class HuiTInputLine;
    class HuiTListBox;
    class HuiTProgressIndicator;
    class HuiTRadioMenu;
    class HuiTStaticText;
    class HuiTTextEdit;
    class HuiTWindow;

    class Widget;
    class GroupWidget;
    class MenuSubmenuWidget;

    class Clickable {
    public:
	/**
	 * Destructor.
	 */
	virtual ~Clickable() {}

	virtual void click() = 0;
    };

    class CommandsToClick {
	typedef map<ushort, Clickable *>	Table;

	Table			mTable;
	ushort			mNewCommand;
	static const ushort	sCommandsBase;

    public:
	/**
	 * Constructor.
	 */
	CommandsToClick()
	    : mNewCommand(sCommandsBase)
	{}

	bool click(ushort aCommand);
	ushort newCommand(Clickable &aClickable);
	void remove(ushort aCommand);
    };

    class HuiTLabel : public TLabel {
    public:
	/**
	 * Constructor.
	 */
	HuiTLabel(const string &aText)
	    : TLabel(TRect(1, 1, 1 + aText.length(), 2), aText.c_str(), 0)
	{
	}

	string
	text() const
	{
	    return(TLabel::text);
	}

	void
	text(const string &aText)
	{
	    delete TLabel::text;
	    TLabel::text = newStr(aText.c_str());
	}

	void
	link(TView *aLink)
	{
	    TLabel::link = aLink;
	}
    };

    class Subwidgets : public list<Widget_P> {
    public:
	/**
	 * Constructor.
	 */
	Subwidgets() {}

	/**
	 * Destructor.
	 */
	~Subwidgets() {}

	Widget_P find(TView *p) const;
	void add(Widget_P w);
	void remove(Widget_P w);
    };

    static inline size_t
    accel_text_length(const string &aText)
    {
	return aText.length() -
		count(aText.begin(), aText.end(), '~') -
		count(aText.begin(), aText.end(), '&');
    }

    static TSItem *
    strings2tsitem(const vector<string> &aTexts)
    {
	TSItem* result = 0;
	for (vector<string>::const_reverse_iterator s = aTexts.rbegin(); s != aTexts.rend(); ++s)
	    result = new TSItem(newStr(HuiTVision::translateAccel(*s).c_str()), result);

	return(result);
    }

    static TPoint
    owner_frame_offset(TView *owner)
    {
	TPoint offset;
	if (owner &&
	    dynamic_cast<TWindow *>(owner) &&
	    dynamic_cast<TWindow *>(owner)->frame) {
	    offset.x = offset.y = 1;
	} else {
	    offset.x = offset.y = 0;
	}
	return(offset);
    }

    class Widget : virtual public Hui::Widget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	Widget_P	mParent;
	TPoint		mPosition;
	int		mFixedSize;

    public:
	/**
	 * Constructor.
	 */
	Widget(Hui &aHui, const string &aName)
	    : Hui::Widget(aHui, aName)
	{
	    setFixedSize(-1);
	}

	/**
	 * Destructor.
	 */
	virtual ~Widget();

	virtual void show();
	virtual void hide();
	virtual bool hidden() const;

	virtual void disable();
	virtual void enable();
	virtual bool disabled() const;

	virtual void move(const Position &aPosition);
	virtual void move(const TPoint &to);
	virtual Position position() const;

	virtual void resize(const Size &aSize);
	virtual void resize(const TPoint &to, int is_arranging = 0);
	virtual Size size() const;

	virtual TView *tview() = 0;
	virtual const TView *tview() const = 0;

	void
	parent(Widget_P aParent)
	{
	    mParent = aParent;
	};

	Widget_P
	parent()
	{
	    return mParent;
	}

	const Widget_P
	parent() const
	{
	    return mParent;
	}

	int
	getFixedSize()
	{
	    return (mFixedSize);
	}

	void
	setFixedSize(int to)
	{
	    mFixedSize = to;
	}

	virtual TRect
	translate(const Position &aPosition, const Size &aSize) const
	{
	    return TRect(aPosition.x(),
			 aPosition.y(),
			 aPosition.x() + aSize.width(),
			 aPosition.y() + aSize.height());
	}

	virtual TPoint
	translate(const Position &aPosition) const
	{
	    TPoint p;
	    p.x = aPosition.x();
	    p.y = aPosition.y();
	    return p;
	}

	virtual TPoint
	translate(const Size &aSize) const
	{
	    TPoint p;
	    p.x = aSize.width();
	    p.y = aSize.height();
	    return p;
	}

	virtual Size
	translate(const TPoint &aSize, int) const
	{
	    return Size(aSize.x, aSize.y);
	}

	virtual Position
	translate(const TPoint &aPosition) const
	{
	    return Position(aPosition.x - 1, aPosition.y - 1);
	}

	virtual Size minSize() const = 0;
	virtual Size maxSize() const = 0;

    };

    class GroupWidget : virtual public Hui::GroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTVision&	mHui;
	Subwidgets	mWidgets;
	int		mArrangement;

    public:
	virtual ~GroupWidget();

	virtual LabelWidget_P
	label(const string &aName,
		const string &aText,
		const Position &aPosition);

	virtual ButtonWidget_P
	button(const string &aName,
		const string &aText,
		CallBackData_P aCallbackData,
		const string &aButtonType,
		const Position &aPosition,
		const string &aComment);

	virtual InputLineWidget_P
	inputLine(const string &aName,
		const string &aInitialText,
		int aMaxLength,
		const string &aValidator,
		const string &aTitle,
		const Position &aPosition,
		const string &aComment);

	virtual CheckBoxesWidget_P
	checkBoxes(const string &aName,
		const vector<string> &aTexts,
		const vector<bool> &aInitialValues,
		const string &aTitle,
		const Position &aPosition,
		const vector<string> &aComments);

	virtual RadioMenuWidget_P
	radioMenu(const string &aName,
		const vector<string> &aTexts,
		const int aInitialValue,
		const string &aTitle,
		const Position &aPosition,
		const vector<string> &aComments);

	virtual ListBoxWidget_P
	listBox(const string &aName,
		const vector<string> &aTexts,
		const int aInitiallySelected,
		const string &aTitle,
		const Position &aPosition,
		const string &aComment);

	virtual TextEditWidget_P
	textEdit(const string &aName,
		const string &aInitialText,
		bool aAllowEdit,
		const string &aTitle,
		const Position &aPosition,
		const string &aComment);

	virtual ProgressIndicatorWidget_P
	progressIndicator(const string &aName,
		float aMin,
		float aMax,
		float aInitial,
		const string &aTitle,
		const Position &aPosition,
		const string &aComment);

	     virtual ContainerWidget_P container(const string &aName,
			const Position &aPosition,
			const string &aAlignment,
			const string &aComment);

	void arrange();
	void arrange(const string &aAlignment, bool arrange_new);

	virtual void current(EmbeddedWidget_P aWidget);
	virtual Widget_P current() const;
	virtual void aboutToChangeCurrent(TView *old, TView *current);

	virtual TGroup *tgroup() = 0;
	virtual const TGroup *tgroup() const = 0;

	virtual Widget_P byName(const string &aName) const;
	virtual void insert(Widget_P aWidget);
	virtual void erase();

	const Subwidgets &
	subwidgets() const
	{
	    return mWidgets;
	}

	Subwidgets &
	subwidgets()
	{
	    return mWidgets;
	}

	virtual HuiTVision &hui() const {return mHui;};

    protected:
	GroupWidget(HuiTVision &aHui,
		const string &aName,
		const Position &aPosition,
		const Size &aSize);

	HuiTVision &hui() { return mHui; }
    };

    class ContainerWidget : virtual public Hui::ContainerWidget, public GroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	TPoint		mPosition;
	TPoint		mSize;
	Widget_P	mParent;
	bool		mHidden;
	bool		mDisabled;
	int		mFixedSize;

    public:
	ContainerWidget(HuiTVision &aHui,
			Widget_P aParent,
			const string &aName);

	virtual ~ContainerWidget();

	// Functions, that need to be virtualized
	void show();
	void hide();
	bool hidden() const;

	void disable();
	void enable();
	bool disabled() const;

	void move(const Position &);
	void move(const TPoint &);
	class Position position() const;

	void resize(const Size &);
	void resize(const TPoint &, int is_arranging = 0);
	class Size size() const;

	void parent(Widget_P aParent) { mParent = aParent; };
	const Widget_P parent() const { return mParent ; }
	Widget_P parent() { return mParent ; }

	// Refer to parent's tgroup()
	virtual TGroup *tgroup();
	const TGroup *tgroup() const;

	int getFixedSize() { return(mFixedSize); }
	void setFixedSize(int to) { mFixedSize = to; }

    protected:
	virtual TRect translate(const Position &aPosition,
				const Size &aSize) const {
	    return TRect(aPosition.x(), aPosition.y(),
			 aPosition.x() + aSize.width(),
			 aPosition.y() + aSize.height());
	}
	virtual TPoint translate(const Position &aPosition) const {
	    TPoint p;

	    p.x = aPosition.x();
	    p.y = aPosition.y();
	    return p;
	}
	virtual TPoint translate(const Size &aSize) const {
	    TPoint p;

	    p.x = aSize.width();
	    p.y = aSize.height();
	    return p;
	}
	virtual Size translate(const TPoint &aSize, int) const {
	    return Size(aSize.x, aSize.y);
	}
	virtual Position translate(const TPoint &aPosition) const {
	    return Position(aPosition.x,
				 aPosition.y);
	}
     };

    class ToplevelWidget : virtual public Hui::ToplevelWidget, public GroupWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

   public:
	virtual ~ToplevelWidget();

	virtual void title(const string &aTitle);
	virtual string title() const;

	virtual TGroup *tgroup();
	virtual const TGroup *tgroup() const;
	virtual TWindow *twindow() = 0;
	virtual const TWindow *twindow() const = 0;

	virtual void resized();

    protected:
	ToplevelWidget(HuiTVision &aHui,
			const string &aName,
			const Position &aPosition,
			const Size &aSize);
     };

    class WindowWidget : public Widget, public ToplevelWidget, public Hui::WindowWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTWindow	*mTWindow;

    public:
	WindowWidget(HuiTVision &aHui,
		    const string &aName,
		    const string &aTitle,
		    const Position &aPosition,
		    const Size &aSize);

	virtual ~WindowWidget();

	virtual void show();
	virtual void hide();

	virtual TView *tview();
	virtual const TView *tview() const;
	virtual TWindow *twindow();
	virtual const TWindow *twindow() const;

	virtual Size
	size() const
	{
	    return HuiTVision::Widget::size() - Size(2, 2);
	};

	virtual Size minSize() const { return size(); }
	virtual Size maxSize() const { return size(); }

	virtual void close(); // user clicked close icon
    };

     class DialogWidget : public Widget, public ToplevelWidget, public Hui::DialogWidget {
	  CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	  HuiTDialog	*mTDialog;
	  string	mResult;
	  bool		mRunning;

       public:
	  DialogWidget(HuiTVision &aHui,
			const string &aName,
			const string &aTitle,
			const Position &aPosition,
			const Size &aSize);

	  virtual ~DialogWidget();

	  virtual const string &result() const;
	  virtual const string &exec();

	  virtual bool running() const { return mRunning; }
	  virtual void endModal(const string &aResult);

	  virtual TView *tview();
	  virtual const TView *tview() const;
	  virtual TWindow *twindow();
	  virtual const TWindow *twindow() const;

	  virtual Size size() const { return HuiTVision::Widget::size() - Size(2, 2); };
	  virtual Size minSize() const { return size(); }
	  virtual Size maxSize() const { return size(); }
     };

     class LabelWidget : public Widget, public Hui::LabelWidget {
	  CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	  HuiTStaticText *mLabel;

       public:
	  LabelWidget(Hui &aHui,
		      Widget_P aParent,
		      const string &aName,
		      const string &aText);
	  virtual ~LabelWidget();

	  virtual void text(const string &aText);
	  virtual string text() const;

	  virtual TView *tview();
	  virtual const TView *tview() const;
	  virtual Size minSize() const { return Size(accel_text_length(text()), 1); }
	  virtual Size maxSize() const { return Size(accel_text_length(text()), 1); }
     };

    class ButtonWidget : public Widget, public Hui::ButtonWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	string		mComment;
	HuiTButton	*mButton;

    public:
	ButtonWidget(HuiTVision &aHui,
		Widget_P aParent,
		const string &aName,
		const string &aText,
		CallBackData_P aCallbackData,
		const string &aButtonType,
		const string &aComment);

	virtual ~ButtonWidget();

	virtual void text(const string &aText);
	virtual string text() const;

	virtual TView *tview();
	virtual const TView *tview() const;
	virtual Size minSize() const { return Size(accel_text_length(text()) + 4, 2); }
	virtual Size maxSize() const { return Size(Size::maxWidth(), 2); }

	virtual void callback_completed(const WidgetCallbackData &aCallbackData);
    };

    class TitledWidget : public Widget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTLabel *mTitle;

    public:
	TitledWidget(Hui &aHui,
		Widget_P aParent,
		const string &aName,
		const string &aTitle);

	virtual ~TitledWidget();

	virtual void title(const string &aTitle);
	virtual string title() const;
	virtual void move(const TPoint &to);
	virtual void resize(const TPoint &to, int is_arranging = 0);

	virtual void
	move(const Position &to)
	{
	    Widget::move(to);
	}

	virtual void
	resize(const Size &to)
	{
	    Widget::resize(to);
	}

	virtual Position position() const;
	virtual Size size() const;

	HuiTLabel *ttitle() { return mTitle; }
	const HuiTLabel *ttitle() const { return mTitle; }
    };

     class InputLineWidget : public TitledWidget, public Hui::InputLineWidget {
	  CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	  HuiTInputLine	*mInputLine;
	  string	mComment;

       public:
	  InputLineWidget(Hui &aHui,
			  Widget_P aParent,
			  const string &aName,
			  const string &aInitialText,
			  int aMaxLength,
			  const string &aValidator,
			  const string &aTitle,
			  const string &aComment);

	  virtual void text(const string &aText);
	  virtual string text() const;

	  virtual void title(const string &aTitle) { TitledWidget::title(aTitle); }
	  virtual string title() const { return TitledWidget::title(); }

	  virtual TView *tview();
	  virtual const TView *tview() const;
	  virtual Size minSize() const { return Size(3, ttitle() ? 2 : 1); }
	  virtual Size maxSize() const { return Size(Size::maxWidth(), ttitle() ? 2 : 1); }
     };

     class ButtonsGroupWidget : public TitledWidget, virtual public Hui::ButtonsGroupWidget {
	  CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	  vector<string> mComments;

     public:
	     ButtonsGroupWidget(Hui &aHui,
				Widget_P aParent,
				const string &aName,
				const string &aTitle,
				const vector<string> &aComments);

	  virtual void disable();
	  virtual void enable();
	  virtual void disable_item(int boxno);
	  virtual void enable_item(int boxno);
	  virtual void title(const string &aTitle) { TitledWidget::title(aTitle); }
	  virtual string title() const { return TitledWidget::title(); }

	  virtual TCluster *tcluster() = 0;
	  virtual const TCluster *tcluster() const = 0;
	  virtual Size minSize() const;
	  virtual Size maxSize() const { return Size(Size::maxWidth(), numberOfItems() + (ttitle() ? 1 : 0)); }
     };

    class CheckBoxesWidget : public ButtonsGroupWidget, public Hui::CheckBoxesWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTCheckBoxes *mCheckBoxes;

    public:
	CheckBoxesWidget(
		Hui &aHui,
		Widget_P aParent,
		const string &aName,
		const vector<string> &aTexts,
		const vector<bool> &aInitialValues,
		const string &aTitle,
		const vector<string> &aComments);

	virtual ~CheckBoxesWidget();

	virtual bool
	state(
		int boxno,
		int aNewState);

	virtual bool state(int boxno) const;
	virtual bool toggle(int boxno);

	virtual void text(int boxno, const string &aText);
	virtual string text(int boxno) const;
	virtual int numberOfItems() const;
	virtual bool item_disabled(int boxno) const;

	virtual TCluster *tcluster();
	virtual const TCluster *tcluster() const;
	virtual TView *tview();
	virtual const TView *tview() const;
    };

    class RadioMenuWidget : public ButtonsGroupWidget, public Hui::RadioMenuWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTRadioMenu *mRadioMenu;

    public:
	RadioMenuWidget(
		Hui &aHui,
		Widget_P aParent,
		const string &aName,
		const vector<string> &aTexts,
		int aInitialValue,
		const string &aTitle,
		const vector<string> &aComments);

	virtual ~RadioMenuWidget();

	virtual void set(int itemno);
	virtual int set() const;
	virtual void text(int boxno, const string &aText);
	virtual string text(int boxno) const;
	virtual int numberOfItems() const;
	virtual bool item_disabled(int boxno) const;

	virtual TCluster *tcluster();
	virtual const TCluster *tcluster() const;
	virtual TView *tview();
	virtual const TView *tview() const;
    };

    class ListBoxWidget : public TitledWidget, public Hui::ListBoxWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTListBox	*mListBox;
	string		mComment;

    public:
	vector<string> mTexts;
	string mTitleText;

	ListBoxWidget(
		Hui &aHui,
		Widget_P aParent,
		const string &aName,
		const vector<string> &aTexts,
		int aInitiallySelected,
		const string &aTitle,
		const string &aComment);

	virtual ~ListBoxWidget();

	virtual int numberOfItems() const;
	virtual string text(int aItemNo) const;
	virtual void text(int aItemNo, const string &aText);
	virtual int select() const;
	virtual void select(int aItemNo);
	virtual void add(int aBefore, const string &aText);
	virtual void remove(int aItemNo);
	virtual void removeAll();

	virtual void set_title(const string &aTitle);
	virtual void title(const string &aTitle);
	virtual string title() const;

	virtual TView *tview();
	virtual const TView *tview() const;

	virtual Size
	minSize() const
	{
	    return Size(5, 3 + (ttitle() ? 1 : 0));
	}

	virtual Size
	maxSize() const
	{
	    return Size(Size::maxWidth(), Size::maxHeight());
	}
    };

    class TextEditWidget : public TitledWidget, public Hui::TextEditWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTTextEdit	*mTextEdit;
	string		mComment;

    public:
	TextEditWidget(
		Hui &aHui,
		Widget_P aParent,
		const string &aName,
		const string &aInitialText,
		bool aAllowEdit,
		const string &aTitle,
		const string &aComment);

	virtual ~TextEditWidget();

	virtual void text(const string &aText);
	virtual string text() const;

	virtual void
	title(const string & aTitle)
	{
	    TitledWidget::title(aTitle);
	}

	virtual string
	title() const
	{
	    return TitledWidget::title();
	}

	virtual TView *tview();
	virtual const TView *tview() const;

	virtual Size
	minSize() const
	{
	    return Size(5, 3 + (ttitle() ? 1 : 0));
	}

	virtual Size
	maxSize() const
	{
	    return Size(Size::maxWidth(), Size::maxHeight());
	}
    };

    class ProgressIndicatorWidget : public TitledWidget, public Hui::ProgressIndicatorWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	Hui			&mHui;
	HuiTProgressIndicator	*mProgressIndicator;
	string			mComment;

    public:
	ProgressIndicatorWidget(
		Hui &aHui,
		Widget_P aParent,
		const string &aName,
		float aMin,
		float aMax,
		float aInitial,
		const string &aTitle,
		const string &aComment);

	virtual ~ProgressIndicatorWidget();

	virtual void set(float aValue);
	virtual float set() const;

	virtual void
	title(const string & aTitle)
	{
	    TitledWidget::title(aTitle);
	}

	virtual string
	title() const
	{
	    return TitledWidget::title();
	}

	virtual TView *tview();
	virtual const TView *tview() const;

	virtual Size
	minSize() const
	{
	    return Size(4, 1 + (ttitle() ? 1 : 0));
	}

	virtual Size
	maxSize() const
	{
	    return Size(Size::maxWidth(), 1 + (ttitle() ? 1 : 0));
	}
    };

    class MenuItemWidget : public Widget, virtual public Hui::MenuItemWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	HuiTVision	 &mHui;
	MenuItemWidget *mParent;

    public:
	MenuItemWidget(
		Hui &aHui,
		const string &aName,
		CallBackData_P aCallbackData,
		MenuItemWidget *aParent);

	virtual void
	move(const Position & aPosition)
	{
	    Hui::MenuItemWidget::move(aPosition);
	}

	virtual Position
	position() const
	{
	    return Hui::MenuItemWidget::position();
	}

	virtual void
	resize(const Size & aSize)
	{
	    Hui::MenuItemWidget::resize(aSize);
	}

	virtual Size
	size() const
	{
	    return Hui::MenuItemWidget::size();
	}

	virtual Size
	minSize() const
	{
	    return Size();
	}

	virtual Size
	maxSize() const
	{
	    return Size();
	}

	virtual TView *
	tview()
	{
	    return 0;
	}

	virtual const TView *
	tview() const
	{
	    return 0;
	}

	virtual bool disabled() const;
	virtual void disable();
	virtual void enable();
	virtual void refresh();

	HuiTVision &
	hui()
	{
	    return mHui;
	}

	virtual TMenuItem *tmenuitem() = 0;
	virtual const TMenuItem *tmenuitem() const = 0;

    protected:
	MenuItemWidget *
	parent()
	{
	    return mParent;
	}

	struct KeyCode {
	    const char *key;
	    unsigned short keyCode;
	};

	static KeyCode sKeyCode[];
	unsigned short key2keyCode(const string &aKey);
     };

     class MenuSeparatorWidget : public MenuItemWidget, public Hui::MenuSeparatorWidget
     {
	  CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	  TMenuItem *mSeparator;

       public:
	  MenuSeparatorWidget(
		    Hui &aHui,
		    const string &aName,
		    MenuItemWidget *aParent);

	  virtual TMenuItem *tmenuitem() { return mSeparator; }
	  virtual const TMenuItem *tmenuitem() const { return mSeparator; }
     };

     class MenuButtonWidget : public MenuItemWidget, public Hui::MenuButtonWidget, public Clickable
     {
	  CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	  TMenuItem	*mButton;

       public:
	  MenuButtonWidget(
		Hui &aHui,
		const string &aName,
		const string &aText,
		CallBackData_P aCallbackData,
		const string &aKey,
		const string &aComment,
		MenuItemWidget *aParent);

	  virtual ~MenuButtonWidget();

	  virtual void key(const string &aKey);
	  virtual void text(const string &aText);
	  virtual string text() const;

	  virtual void click() { Hui::MenuButtonWidget::click(); }

	  virtual TMenuItem *tmenuitem() { return mButton; }
	  virtual const TMenuItem *tmenuitem() const { return mButton; }
    };

    class MenuSubmenuWidget : public MenuItemWidget, public Hui::MenuSubmenuWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	TSubMenu *mSubMenu;

    public:
	MenuSubmenuWidget(
		Hui &aHui,
		const string &aName,
		const string &aText,
		const string &aComment,
		MenuItemWidget *aParent);

	virtual MenuItemWidget_P
	addButton(
		const string &aName,
		const string &aText,
		CallBackData_P aCallbackData,
		const string &aKey,
		const string &aComment,
		MenuItemWidget_P aBefore);

	virtual MenuItemWidget_P
	addSeparator(
		const string &aName,
		MenuItemWidget_P aBefore);

	virtual MenuItemWidget_P
	addSubmenu(
		const string &aName,
		const string &aText,
		const string &aComment,
		MenuItemWidget_P aBefore);

	virtual void
	add(
		MenuItemWidget_P aBefore,
		MenuItemWidget_P aItem);

	virtual void remove(MenuItemWidget_P aItem);
	virtual void text(const string &aText);
	virtual string text() const;

	virtual TMenuItem *tmenuitem();
	virtual const TMenuItem *tmenuitem() const;
	TSubMenu *&submenu() { return mSubMenu; }
    };

    class MenuBarWidget : public MenuSubmenuWidget {
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	TMenuBar *mMenuBar;

    public:
	MenuBarWidget(Hui &aHui);

	MenuItemWidget_P
	addSubmenu(
		const string &aName,
		const string &aText,
		const string &aComment,
		MenuItemWidget_P aBefore);

	void refresh();
    };

    const Subwidgets &
    subwidgets() const
    {
	return mWidgets;
    }

    Subwidgets &
    subwidgets()
    {
	return mWidgets;
    }

    CommandsToClick &
    commandsToClick()
    {
	return mCommandsToClick;
    }

    static string translateAccel(const string &aText);

    virtual void handleEvent(TEvent &event);

    /*
     * Variables
     */
    Subwidgets			mWidgets;
    MenuSubmenuWidget_P		mMenuBar;
    CommandsToClick		mCommandsToClick;

public:
    /**
     * Constructor.
     */
     HuiTVision()
	: TProgInit(&HuiTVision::initStatusLine,
		    &HuiTVision::initMenuBar,
		    &HuiTVision::initDeskTop)
		//, mCommands(0), mButtonCommands(0), mWidgets(0)
    {
    }

    /**
     * Run the application.
     */
    void run();

    /**
     * Create a widget by name.
     */
    Widget_P
    byName(const string &aName);

    /**
     * Create a window.
     */
    WindowWidget_P
    window(const string &aName,
	   const string &aTitle,
	   const Position &aPosition,
	   const Size &aSize,
	   const string &aAlignment,
	   bool no_menu = false);

    /**
     * Create a dialog.
     */
    DialogWidget_P
    dialog(Widget_P aParent,
	   const string &aName,
	   const string &aTitle,
	   const Position &aPosition,
	   const Size &aSize,
	   const string &aAlignment);

    /**
     * Create a menu bar.
     */
    MenuSubmenuWidget_P
    menuBar();

    /**
     */
    void processPendingEvents();

    /**
     * Returns an empty string if user pressed cancel button.
     */
    string
    fileSelectionDialog(const string &aTitle, const string &aWildCard);

    /**
     * Returns type of UI.
     */
    const char *type() const { return "text"; }

    /**
     * Returns size of UI.
     */
    Size size() const;

    /**
     * Exit the application.
     */
    void exit(int status);

    /**
     * Destructor.
     */
    ~HuiTVision() {}
};

#endif // HuiTVision_hh
