/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/include/sysinstall/Feature.hh,v 1.19 2002/07/01 20:59:47 antoine Exp $
 */

// -*- C++ -*-  Fri Aug  7 1998
// $Id: Feature.hh,v 1.6 1999/01/21 11:07:19 eugene Exp $

#ifndef Feature_hh
#define Feature_hh

#include "LanguageInterface.hh"
#include "Container.hh"

#include <vector>
#include <stdexcept>
#include <list>
#include <ctime>

class Feature : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     Feature() : mName ( "nil" ), mVersion ( "0" ), mSerial( 0 ), mSnapshot( 0 ) {}
     Feature( const string& aName,
              const string& aVersion,
              unsigned aSerial, 
              time_t aSnapshot ) : 
          mName( aName ),
          mVersion( aVersion ), 
          mSerial( aSerial ), 
          mSnapshot( aSnapshot ) {}

     const string& name() const { return mName; }
     const string& version() const { return mVersion.to_string(); }
     unsigned serial() const { return mSerial; }
     time_t snapshot() const { return mSnapshot; }

     enum Comparison
     {
          cEqual = 0,           /// this is equal
          cThisOlder = -1,      /// this is older
          cThisNewer = 1,       /// this is newer
     };

     Comparison compare( const Feature& toCompare ) const;
     Comparison compare( Pointer<Feature> toCompare ) const 
          { return compare( *toCompare ); }
     bool operator == ( const Feature& toCompare ) const 
          { return name() == toCompare.name() && compare( toCompare ) == cEqual; }
     bool operator == ( Pointer<Feature> toCompare ) const 
          { return operator == ( *toCompare ); }

     virtual string print() const;
     /// return number of characters scanned
     virtual size_t scan( const char* text ) throw (invalid_argument);

     static Pointer<Feature> make( const string& aName,
                                   const string& aVersion, 
                                   unsigned aSerial,
                                   time_t aSnapshot )
          {
               if ( aSnapshot == time_t( -1 ) )
                    aSnapshot = ::time( 0 );
               return new Feature( aName, aVersion, aSerial, aSnapshot );
          }

     static inline bool lessByName( Pointer<Feature> a,
                                    Pointer<Feature> b )
          {
               return a->name() < b->name();
          }

     static inline bool equalByName( Pointer<Feature> a,
                                     Pointer<Feature> b )
          {
               return a->name() == b->name();
          }

     class Version {

     public:
          Version() : mData("0") {}
          Version(const string& data) 
               throw (invalid_argument) : mData(data)  {
               if (!consistent())
                    throw invalid_argument("invalid version string");
          }
          const string& to_string() const { return mData; }
          Comparison compare(const Version& toCompare ) const;
          bool operator == ( const Version& toCompare ) const
               { return compare( toCompare ) == cEqual; }
          bool operator > ( const Version& toCompare ) const
               { return compare( toCompare ) == cThisNewer; }

          operator string() const { return to_string(); }          

          static void tokenize(const string& str,
                               vector<string>& tokens,
                               const string& delimiters);


     private:
          /**
           * checks for version string consistency
           *
           * a version string is valid if it matches the regexp
           * ^[0-9][0-9]*(\.[0-9][0-9]*)*$
           */
          bool consistent() const;
          string mData;

     };

  private:
     string mName;
     Version mVersion;
     unsigned mSerial;
     time_t mSnapshot;

     /// Compares version strings.
     Comparison compare_versions( const Feature& toCompare ) const;
     Comparison compare_serials( const Feature& toCompare ) const;
     Comparison compare_snapshots( const Feature& toCompare ) const;

}; // class Feature

typedef Container<list<Pointer<Feature> > > FeatureList;

#endif // Feature_hh
