#ifndef Dim_hh_9c0d91a4
#define Dim_hh_9c0d91a4
/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/Hui/Dim.hh,v 1.7 2002/08/26 11:05:31 okumoto Exp $
 */

#include <iostream>
#include "limits"

typedef short			Dimension;

struct Position {
	Dimension	x;
	Dimension	y;

	/**
	 * Constructor.
	 */
	Position(Dimension aX = 0, Dimension aY = 0)
	    : x(aX), y(aY)
	{
	}

	inline Position &
	operator+=(const Position &rhs)
	{
	    x += rhs.x;
	    y += rhs.y;
	    return *this;
	}

	inline const Position
	operator+(const Position &rhs) const
	{
	    return Position(*this) += rhs;
	}

	inline Position &
	operator-=(const Position &rhs)
	{
	    x -= rhs.x;
	    y -= rhs.y;
	    return *this;
	}

	inline const Position
	operator-(const Position &rhs) const
	{
	    return Position(*this) -= rhs;
	}
};

struct Size {
	Dimension	width;
	Dimension	height;

	/**
	 * Constructor.
	 */
	Size(Dimension aWidth = 0, Dimension aHeight = 0)
	    : width(aWidth), height(aHeight)
	{
	}

	inline static Dimension
	maxWidth()
	{
	    return numeric_limits<Dimension>::max();
	}

	inline static Dimension
	maxHeight()
	{
	    return numeric_limits<Dimension>::max();
	}

	Size &
	operator+=(const Size &rhs)
	{
	    if (width != maxWidth() && rhs.width != maxWidth())
		width = width + rhs.width;
	    else
		width = maxWidth();

	    if (height != maxHeight() && rhs.height != maxHeight())
		height = height + rhs.height;
	    else
		height = maxHeight();
	    return *this;
	}

	Size &
	operator+=(const Position &rhs)
	{
	    if (width != maxWidth())
		width = width + rhs.x;
	    if (height != maxHeight())
		height = height + rhs.y;
	    return *this;
	}

	Size
	operator+(const Size &rhs) const
	{
	    Size result = *this;
	    result += rhs;
	    return result;
	}

	Size
	operator+(const Position &rhs) const
	{
	    Size result = *this;
	    result += rhs;
	    return result;
	}

	Size &
	operator-=(const Size &rhs)
	{
	    if (width != maxWidth())
		width = width - rhs.width;
	    if (height != maxHeight())
		height = height - rhs.height;
	    return *this;
	}

	Size &
	operator-=(const Position &rhs)
	{
	    if (width != maxWidth())
		width = width - rhs.x;
	    if (height != maxHeight())
		height = height - rhs.y;
	    return *this;
	}

	Size
	operator-(const Size &rhs) const
	{
	    Size result = *this;
	    result -= rhs;
	    return result;
	}

	Size
	operator-(const Position &rhs) const
	{
	    Size result = *this;
	    result -= rhs;
	    return result;
	}
};

/**
 * Represents a region on the UI. The upper left corner is inside
 * the box, while the bottom right corner is outside.
 */
class HuiRegion {
public:
	Position	a;	/**< top left corner */
	Position	b;	/**< bottom right corner */

	/**
	 * Constructor.
	 *
	 * @param aSize		Size of region
	 * @param tl		Top left corner
	 */
	HuiRegion(const Size &aSize, const Position &tl)
	    : a(tl), b(tl.x + aSize.width, tl.y + aSize.height)
	{
	}

	/**
	 * Constructor.
	 *
	 * @param width		width of region
	 * @param height	hight of region
	 * @param ax		x coordinate of upper left corner
	 * @param ay		y coordinate of upper left corner
	 */
	HuiRegion(Dimension width = 0, Dimension height = 0,
		  Dimension ax = 0, Dimension ay = 0)
	    : a(ax, ay), b(ax + width, ay + height)
	{
	}

	/**
	 * Constructor.
	 *
	 * @param tl		Top left corner
	 * @param br		Bottom right corner
	 */
	HuiRegion(const Position &tl, const Position &br)
	    : a(tl), b(br)
	{
	}

	/**
	 * @param width		width of region
	 * @param height	hight of region
	 * @param ax		x coordinate of upper left corner
	 * @param ay		y coordinate of upper left corner
	 */
	inline void
	set(Dimension width = 0, Dimension height = 0,
	    Dimension ax = 0, Dimension ay = 0)
	{
		a = Position(ax, ay);
		b = Position(ax + width, ay + height);
	}

	inline Size
	size() const
	{
		return Size((b.x - a.x), (b.y - a.y));
	}

	inline Dimension
	width() const
	{
		return b.x - a.x;
	}

	inline Dimension
	height() const
	{
		return b.y - a.y;
	}

	/**
	 * Make a larger version.
	 */
	HuiRegion
	bigger(int s) const
	{
		return HuiRegion(
			Size(s * width(), s * height()),
			Position(s * a.x, s * a.y));
	}
};

#endif /* Dim_hh_9c0d91a4 */
