/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/HuiQt/HuiQtTextEdit.cc,v 1.1 2002/08/21 10:52:07 okumoto Exp $
 */

#include "HuiQt/HuiQtTextEdit.hh"

#define Uses_TEditor
#define Uses_TEvent
#define Uses_TGroup
#define Uses_TKeys
#define Uses_TScrollBar

#include <tvision/tv.h>

/**
 */
class HuiQtTEditor : public TGroup {
	/**
	 * TEditor with dark gray text and light gray background.
	 */
	class GrayTEditor : public TEditor {
	public:
		GrayTEditor(
			const TRect& bounds,
			TScrollBar *hor,
			TScrollBar *ver,
			TIndicator *ind,
			uint size)
		    : TEditor(bounds, hor, ver, ind, size)
		{
		}

		/**
		 * Set the color of the text.
		 */
		TPalette &
		getPalette() const
		{
#define cpHuiTTextEdit "\x1A\x19"
			static TPalette palette(
					cpHuiTTextEdit,
					sizeof(cpHuiTTextEdit) - 1);
			return palette;
		}

		void
		handleEvent(TEvent &event)
		{
			if ((event.what == evKeyDown) &&
			    (event.keyDown.keyCode == kbTab)) {
				TView::handleEvent(event);
			} else {
				TEditor::handleEvent(event);
			}
		}
	};

	TScrollBar	hsb;
	TScrollBar	vsb;
	GrayTEditor	editor;

public:
	/**
	 * Construct a text editor with scroll bars.
	 *
	 * @param bounds	Position and size of editor.
	 * @param aBufSize	Size of edit buffer.
	 */
	HuiQtTEditor(
		const HuiRegion &region,
		uint aBufSize,
		bool readonly)
	    : TGroup(TRect(
			region.a.x, region.a.y,
			region.b.x, region.b.y)),
	      hsb(TRect(
			0, region.height() - 1,
			region.width() - 1, region.height())),
	      vsb(TRect(
			region.width() - 1, 0,
			region.width(), region.height() - 1)),
	      editor(TRect(
			0, 0,
			region.width() - 1, region.height() - 1),
		&hsb, &vsb, NULL, aBufSize)
	{
		insert(&hsb);
		insert(&vsb);
		insert(&editor);
	}

	/**
	 * Destructor.
	 */
	~HuiQtTEditor()
	{
		shutDown();
	}

	/**
	 * Releases resources.  We don't want to destroy the list
	 * of sub views since we they are part of this object.
	 */
	void
	shutDown()
	{
		hsb.shutDown();
		vsb.shutDown();
		editor.shutDown();
	}

	void
	text(const string &aText)
	{
		editor.deleteRange(0, editor.bufLen, False);

		editor.insertText(
			aText.c_str(),
			aText.length(),
			False);

		editor.setCurPtr(0, 0);

		editor.drawView();
	}

	string
	text() const
	{
		return string(editor.buffer, editor.bufLen);
	}
};

/**
 * Constructor.
 */
HuiQtTextEdit::HuiQtTextEdit(
	HuiWidget *aParent,
	const string &aText,
	bool aAllowEdit,
	const string &aTitle,
	const HuiRegion &region)
    : HuiQtEmbedded(aParent)
{
	HuiQtTEditor *editor = 
		new HuiQtTEditor(region, 1024 * 100, aAllowEdit);

	editor->text(aText);
	mView = editor;
}

/**
 * Destructor.
 */
HuiQtTextEdit::~HuiQtTextEdit()
{
	delete mView;
}

void
HuiQtTextEdit::text(const string &aText)
{
	HuiQtTEditor *editor = dynamic_cast<HuiQtTEditor *>(mView);
	return editor->text(aText);
}

string
HuiQtTextEdit::text() const
{
	HuiQtTEditor *editor = dynamic_cast<HuiQtTEditor *>(mView);
	return editor->text();
}

