/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/HuiTv/HuiTvInputLine.cc,v 1.38 2002/08/20 10:41:42 okumoto Exp $
 */

#include "HuiTv/HuiTvInputLine.hh"
#include "HuiTv/HuiTvTitle.hh"

#define Uses_MsgBox
#define Uses_TFilterValidator
#define Uses_TInputLine
#define Uses_TPXPictureValidator
#define Uses_TParamText
#define Uses_TRangeValidator
#define Uses_TRect

#include <tvision/tv.h>

extern "C" {
	#include <stdio.h>	/* sscanf() */
}

class HuiTInputLine : public TInputLine {
	class TIntegerRangeValidator : public TRangeValidator {
	public:
		TIntegerRangeValidator(long aMin, long aMax)
		    : TRangeValidator(aMin, aMax)
		{
		}

		void
		error()
		{
#if 0
			messageBox(mfError | mfOKButton,
				_("Value not in the range %ld to %ld"),
				min, max);
#else
			assert(0);
#endif
		}
	};

	class TRealRangeValidator : public TFilterValidator {
		double min;
		double max;
	public:
		TRealRangeValidator(double aMin, double aMax)
			: TFilterValidator(NULL),
			  min(aMin),
			  max(aMax)
		{
			validChars = newStr("0123456789+-eE.");
		}

		void
		error()
		{
#if 0
			messageBox(mfError | mfOKButton,
				_("Value not in the range %lg to %lg"),
				min, max);
#else
			assert(0);
#endif
		}

		Boolean
		isValid(const char s[])
		{
			if (TFilterValidator::isValid(s) == False) {
				return(False);
			}

			/* check if input is parsable as a real */
			double value;
			if (sscanf(s, "%le", &value) != 1) {
				return(False);
			}

			/* check if input is within specified range */
			if ((value < min) || (value > max)) {
				return(False);
			}

			return(True);
		}
	};

public:
	/**
	 * Construct.
	 */
	HuiTInputLine(
		const string &aText,
		int aMaxLength,
		const string &aValidator)
	    : TInputLine(TRect(0, 0, 10, 1), aMaxLength, NULL)
	{
		setData(const_cast<char *>(aText.c_str()));
		setValidator(aValidator);
	}

	void
	setValidator(const string &aValidator)
	{
		TValidator* validator = NULL;
		if (aValidator == "none" || aValidator == "") {
			validator = NULL;

		} else if (!strncmp(aValidator.c_str(), "picture", 7)) {
			validator = new TPXPictureValidator(
					aValidator.c_str() + 8,
					True);

		} else if (!strncmp(aValidator.c_str(), "valid_chars", 11)) {
			validator = new TFilterValidator(
					aValidator.c_str() + 12);

		} else if (!strncmp(aValidator.c_str(), "integer_range", 13)) {
			long min = 0;
			long max = 0;
			if (sscanf(aValidator.c_str(), "integer_range %ld %ld", &min, &max) != 2) {
				cerr << "Warning: Invalid integer_range "
					"validator argument, -Inf +Inf "
					"will be used\n";
				min = numeric_limits<long>::min();
				max = numeric_limits<long>::max();
			}
			validator = new TIntegerRangeValidator(min, max);

		} else if (!strncmp(aValidator.c_str(), "real_range", 10)) {
			double min = 0.0;
			double max = 0.0;
			if (sscanf(aValidator.c_str(), "real_range %lf %lf", &min, &max) != 2) {
				cerr << "Warning: Invalid real_range "
					"validator argument, -Inf +Inf "
					"will be used\n";
				min = numeric_limits<double>::min();
				max = numeric_limits<double>::max();
			}
			validator = new TRealRangeValidator(min, max);
		} else {
			cerr << "Warning: Unknown validator for "
				"InputLineWidget: " << aValidator <<
				" (\"none\" will be used)" << endl;
			validator = NULL;
		}

		TInputLine::setValidator(validator);
	}
};

HuiTvInputLine::HuiTvInputLine(
	HuiWidget *aParent,
	const string &aInitialText,
	int aMaxLength,
	const string &aValidator,
	const string &aTitle)
    : HuiTvEmbedded(aParent)
{
	if (aTitle.empty()) {
		mView = new HuiTInputLine(aInitialText, aMaxLength, aValidator);
	} else {
		mView = new HuiTvTitle(
				aTitle,
				new HuiTInputLine(
					aInitialText,
					aMaxLength,
					aValidator),
				TRect(0, 0, 10, 2));
	}
}

void
HuiTvInputLine::text(const string &aText)
{
	TInputLine *widget = dynamic_cast<TInputLine *>(mView);

	widget->setData(const_cast<char *>(aText.c_str()));
}

string
HuiTvInputLine::text() const
{
	TInputLine *widget = dynamic_cast<TInputLine *>(mView);

	char s[widget->dataSize() + 1];
	widget->getData(s);

	return s;
}

