/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/HuiTv/HuiTvWindow.cc,v 1.38 2002/08/20 10:41:42 okumoto Exp $
 */

#include "Hui/Dim.hh"
#include "HuiTv/HuiTv.hh"
#include "HuiTv/HuiTvMenuBar.hh"
#include "HuiTv/HuiTvWindow.hh"

#define Uses_TDialog
#define Uses_TEvent
#define Uses_TPalette
#define Uses_TWindow
#include <tvision/tv.h>

/**
 * Represent a window which can contain a menu bar.
 */
class MyTWindow : public TWindow {
	HuiTvWindow	&owner;
	TView		*mCurrent;
	HuiTvMenuBar	*mMenuBar;

public:
	/**
	 * Construct a TVision window.  The window created is
	 * not numbered (wnNoNumber), so you can not use any of the
	 * ALlt-n keys to select this window.
	 *
	 * @param region	Location and Size of window
	 * @param aTitle	Title displayed for the window.
	 */
	MyTWindow(
		HuiTvWindow &aWindow,
		const HuiRegion &region,
		const string &aTitle,
		short aNumber)
	    : TWindow(
			TRect(
				region.a.x,
				region.a.y,
				region.b.x,
				region.b.y),
			aTitle.c_str(),
			wnNoNumber),
	      TWindowInit(&TWindow::initFrame),
	      owner(aWindow),
	      mCurrent(NULL),
	      mMenuBar(NULL)
	{
		setState(sfShadow, False);	// no shadow
	}

	/**
	 * Return color of window.
	 */
	TPalette &
	getPalette() const
	{
		static TPalette paletteCyan(cpCyanDialog,
					    sizeof(cpCyanDialog) - 1);
		return paletteCyan;
	}

	/**
	 */
	void
	handleEvent(TEvent &event)
	{
		if (event.what == evCommand &&
		    event.message.command == cmClose) {
			clearEvent(event);
			owner.close();
			return;
		}

		bool resize_event = (event.what == evCommand) &&
				    ((event.message.command == cmResize) ||
				     (event.message.command == cmZoom));

		TWindow::handleEvent(event);

		if (mCurrent != current) {
			owner.aboutToChangeCurrent(mCurrent, current);
			mCurrent = current;
		}
		if (resize_event) {
			owner.resized();
		}
	}

	/**
	 * Return a pointer to the menu bar.  If none exists, create
	 * one and return a pointer to it.
	 */
	HuiMenuSubmenu *
	menuBar()
	{
		if (mMenuBar == NULL) {
			/*
			 * Create and add the menu bar.
			 */
			TRect bounds = getExtent();
			bounds.a.x++;
			bounds.a.y++;
			bounds.b.x--;
			bounds.b.y = bounds.a.y + 1;

			mMenuBar = new HuiTvMenuBar(bounds);

			insert(mMenuBar->widget());
		}
		return mMenuBar;
	}
};

/**
 * Construct a top level window.
 */
HuiTvWindow::HuiTvWindow(
	const string &aTitle,
	const string &aAlignment,
	const HuiRegion &region)
  : HuiTvToplevel()
{
	mView = new MyTWindow(*this, region, aTitle, wnNoNumber);
}

/**
 * Destructor.
 */
HuiTvWindow::~HuiTvWindow()
{
	MyTWindow *window = dynamic_cast<MyTWindow *>(mView);

	aboutToChangeCurrent(window->current, NULL);

	destroy(window);
}

/**
 * Return a pointer to the menu bar.
 */
HuiMenuSubmenu *
HuiTvWindow::menuBar()
{
	MyTWindow *window = dynamic_cast<MyTWindow *>(mView);

	return window->menuBar();
}

