/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/hui/HuiQt_ListBoxWidget.cc,v 1.36 2002/05/27 20:22:46 alex Exp $
 */

// -*- C++ -*-   Wed Dec  2 1998
// Eugene Skepner <eu@qub.com>
// $Id: HuiQt.cc,v 1.1.1.1 2000/02/15 16:30:35 alex Exp $

#include "hui/HuiQt.hh"

#include <qlabel.h>

#include "String.hh"

#define MAX(a,b) ((a) > (b) ? (a) : (b))

//======================================================================

HuiQt::ListBoxWidget::ListBoxWidget(Hui& aHui,
				    Pointer<GroupWidget> aParent,
				    const string& aName,
				    const vector<string>& aTexts,
				    int aInitiallySelected,
				    const string& aTitle,
				    const string& aComment)
	: Hui::Widget(aHui, aName),
	  Hui::EmbeddedWidget(aHui, aName),
	  TitledWidget(aHui, aParent, aName, string()),
	  Hui::ListBoxWidget(aHui, aName),
	  mTexts(aTexts),
	  mTitleText(aTitle),
	  mTopRow(NULL),
	  mComment(aComment)
{
	parent(aParent);
	
	ContainerWidget *cp = dynamic_cast<ContainerWidget*>((Hui::Widget*) parent());
	GroupWidget *gp = dynamic_cast<GroupWidget*>((Hui::Widget*) parent());

	if (cp != NULL)
		mListBox = new QListViewEnterLeave(const_cast<QWidget*>(cp->qwidget()),
						  aName.c_str());
	else if (gp != NULL)
		mListBox = new QListViewEnterLeave(const_cast<QWidget*>(gp->qwidget()),
						  aName.c_str());
	
	mListBox->setSelectionMode(QListView::Single);
	mListBox->setAllColumnsShowFocus(true);
	mListBox->setSorting(-1);
	
	updatetitle(aTitle);
	updatetexts();

	mListBox->resize(mListBox->sizeHint());
	mListBox->setMinimumSize(mListBox->sizeHint());

	setsize(mListBox->size());

	select(aInitiallySelected);
	mListBox->repaint();

	mListBox->connect(mListBox,
			  SIGNAL(enter_signal()),
			  &huiQtEnterLeave(),
			  SLOT(enter()));
	mListBox->connect(mListBox,
			  SIGNAL(leave_signal()),
			  &huiQtEnterLeave(),
			  SLOT(leave()));

} // HuiQt::ListBoxWidget::ListBoxWidget

//======================================================================

void
HuiQt::ListBoxWidget::updatetexts()
{
	int selected = select();
	int ncols = count_columns(mTexts, mTitleText);
	int nrows = mTexts.size();
	QListViewItem *node = NULL, *oldnode;

	select(selected);

	if (mListBox != NULL) {
		int x, y;

		mListBox->clear();
		mTopRow = NULL;

		for (y = 0; y < nrows; y++) {
			oldnode = node;
			node = new QListViewItem(mListBox, oldnode);
			for (x = 0; x < ncols; x++) {
				string cell = extract_cell(mTexts, x, y);
				node->setText(x, QString(cell.c_str()));
			}
			mListBox->insertItem(node);
			if (y == 0)
				mTopRow = node;
		}


	}

	select(selected);
}

//======================================================================

int
HuiQt::ListBoxWidget::count_columns(const vector<string>& aTexts,
				    const string& aTitle)
{
	int max = 1;
	string::size_type tmp;
	size_t item;
	int here;

	for (item = 0; item < aTexts.size(); item++) {
		tmp = string::npos;
		here = 1;
		while((tmp = aTexts[item].find("\t", tmp + 1)) != aTexts[item].npos)
			here++;
		max = MAX(max, here);
	}

	tmp = aTitle.npos;
	here = 1;
	while((tmp = aTitle.find("\t", tmp + 1)) != aTitle.npos)
		here++;
	max = MAX(max, here);

	return max;
}

//======================================================================

string
HuiQt::ListBoxWidget::extract_cell(const vector<string>& aTexts, int x, int y)
{
	string::size_type tmp = string::npos;
	string::size_type next = string::npos;
	int where = 0;


	if ((unsigned) y >= aTexts.size())
		return string("");	

	string row = aTexts[y];
	while(where < x && (tmp = row.find("\t", tmp + 1)) != row.npos)
		where++;
	if (where < x)
		return string("");	
	
	next = row.find("\t", tmp + 1);
	if (next == row.npos)
		next = row.length();
	else
		next--;
	return row.substr(tmp + 1, next - tmp);
		
}

//======================================================================

void
HuiQt::ListBoxWidget::updatetitle(const string &aTitle)
{
	string::size_type tmp = string::npos;
	string::size_type next = string::npos;
	string col;
	int ncols = count_columns(mTexts, aTitle);
	int where = -1;

	tmp = 0;
	do {
		where++;
		next = aTitle.find("\t", tmp + 1);
		if (next == aTitle.npos)
			next = aTitle.length();
		else
			next--;
		col = aTitle.substr(tmp, next - tmp + 1);
		if (ncols > mListBox->columns())
			mListBox->addColumn(QString(col.c_str()));
		else
			mListBox->setColumnText(where, QString(col.c_str()));
		tmp = next + 2;
	} while (next < aTitle.length());

	while (++where < ncols) {
		mListBox->setColumnText(where, QString(""));
	}

	while (ncols != mListBox->columns()) {
		if (ncols > mListBox->columns())
			mListBox->addColumn(QString(""));
		else
			mListBox->removeColumn(mListBox->columns() - 1);
	}

}

//======================================================================

int
HuiQt::ListBoxWidget::numberOfItems() const
{
	if ( mListBox )
		return mTexts.size();
	else
		return 0;

} // HuiQt::ListBoxWidget::numberOfItems

//======================================================================

string
HuiQt::ListBoxWidget::text(int aItemNo) const
{
	if (mListBox) {
		if (size_t(aItemNo) >= mTexts.size())
			throw runtime_error(string("ListBox itemno ")
					    + to_string(aItemNo)
					    + " is out of range 0.."
					    + to_string(mTexts.size() - 1));
		return mTexts[aItemNo];
	}
	else
		return string();
	
} // HuiQt::ListBoxWidget::text

//======================================================================

void
HuiQt::ListBoxWidget::text(int aItemNo, const string& aText)
{
	if (mListBox) {
		if (size_t(aItemNo) >= mTexts.size())
			throw runtime_error(string("ListBox itemno " )
					    + to_string(aItemNo)
					    + " is out of range 0.."
					    + to_string(mTexts.size() - 1));
		mTexts[aItemNo] = aText;
		updatetitle(mTitleText);
		updatetexts();
	}
	
} // HuiQt::ListBoxWidget::text

//======================================================================

int
HuiQt::ListBoxWidget::select() const
{

	if (mListBox) {
		QListViewItem *rowptr = mTopRow;
		int i = 0;

		while (rowptr != NULL && !rowptr->isSelected()) {
			rowptr = rowptr->itemBelow();
			i++;
		}
		return i;
	} else
		return 0;

} // HuiQt::ListBoxWidget::select

//======================================================================

void
HuiQt::ListBoxWidget::select(int aItemNo)
{
	if (aItemNo == -1)
		return;

	if (numberOfItems() && mListBox) {
		int i = 0;
		QListViewItem *rowptr = mTopRow;

		if (size_t(aItemNo) >= mTexts.size())
			throw runtime_error(string("ListBox itemno ")
					    + to_string(aItemNo)
					    + " is out of range 0.."
					    + to_string(mTexts.size() - 1));

		while (i < aItemNo) {
			if (rowptr != NULL) {
				rowptr = rowptr->itemBelow();
				i++;
			}
		}
		if (rowptr != NULL)
			mListBox->setSelected(rowptr, TRUE);
	}
	
} // HuiQt::ListBoxWidget::select

//======================================================================

void
HuiQt::ListBoxWidget::add(int aBefore, const string& aText)
{
	int selected = select();

	if (aBefore < 0 || aBefore > numberOfItems())
		throw runtime_error(string("ListBox beforeItem ")
				    + to_string(aBefore)
				    + " is out of range 0.."
				    + to_string(mTexts.size()));
	
	mTexts.insert(mTexts.begin() + aBefore, aText);
		
	if (mListBox) {
		if (aBefore <= select()) {
			/* Insert dummy one first to keep correct selection! */
			new QListViewItem(mListBox, mTopRow);
		}
		updatetitle(mTitleText);
		updatetexts();
	}

	/* workaround if old selected one was first entry */
	if (aBefore == 0 && selected == 0 && numberOfItems() > 1)
		select(1);
		

} // HuiQt::ListBoxWidget::add

//======================================================================

void
HuiQt::ListBoxWidget::remove(int aItemNo)
{
	mTexts.erase(mTexts.begin() + aItemNo);
	
	if (mListBox) {
		if (aItemNo > 0 && select() >= aItemNo)
			select(select() - 1);
		updatetitle(mTitleText);
		updatetexts();
	}

} // HuiQt::ListBoxWidget::remove

//======================================================================

void
HuiQt::ListBoxWidget::removeAll()
{
	if (mListBox)
		mListBox->clear();
	mTexts.clear();
	
} // HuiQt::ListBoxWidget::removeAll

//======================================================================

QWidget*
HuiQt::ListBoxWidget::qwidget()
{
	return(mListBox);

} // HuiQt::ListBoxWidget::qwidget

//======================================================================

const QWidget*
HuiQt::ListBoxWidget::qwidget() const
{
	return(mListBox);
	
} // HuiQt::ListBoxWidget::qwidget

//======================================================================
