/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/Feature.cc,v 1.23 2002/07/01 17:04:25 antoine Exp $
 */

// -*- C++ -*-   Fri Aug  7 1998
// $Id: Feature.cc,v 1.3 1998/12/31 13:31:21 eugene Exp $

#include "sysinstall/Feature.hh"
#include "String.hh"
#include "Log.hh"

#include <sstream>
#include <cctype>
#include <cstdlib>

//======================================================================

Feature::Comparison Feature::compare( const Feature& toCompare ) const
{
     if ( name() != toCompare.name() )
          throw invalid_argument( "Cannot compare features with different names" );

     const Comparison vercmp = compare_versions( toCompare );
     const Comparison sercmp = compare_serials( toCompare );
     const Comparison snapcmp = compare_snapshots( toCompare );

     if ( sercmp == cEqual )
          if ( snapcmp == cEqual )
               return vercmp;
          else return snapcmp;
     else return sercmp;
} // Feature::compare

//======================================================================

Feature::Comparison Feature::compare_versions( const Feature& toCompare ) const
{
        return mVersion.compare(toCompare.version());
} // Feature::compare_versions

//======================================================================

Feature::Comparison Feature::compare_serials( const Feature& toCompare ) const
{
     if ( serial() > toCompare.serial() )
          return cThisNewer;
     if ( serial() < toCompare.serial() )
          return cThisOlder;
     return cEqual;
} // Feature::compare_serials

//======================================================================

Feature::Comparison Feature::compare_snapshots( const Feature& toCompare ) const
{
     if ( snapshot() > toCompare.snapshot() )
          return cThisNewer;
     if ( snapshot() < toCompare.snapshot() )
          return cThisOlder;
     return cEqual;
} // Feature::compare_snapshots

//======================================================================

string Feature::print() const
{
     return name() + " " + version() + " " + to_string( serial() ) + " " + to_string( snapshot() );
} // Feature::print

//======================================================================

/*
 * XXX: this whole thing should be scrapped in favor in a stream-based
 * approach
 */
size_t Feature::scan( const char* text ) throw (invalid_argument)
{
     istringstream ist(text);
     string item[4];
     size_t length  = 0;
     
     for (int i = 0; i < 4; i++) {
          if ( !(ist >> item[i]) )
               throw invalid_argument( string( "not enough arguments: " ) + text );
          length += item[i].length();
     }

     mName = item[0];
     mVersion = item[1];
     mSerial = from_string( item[2], mSerial );
     mSnapshot = from_string( item[3], mSnapshot );

     return length;
} // Feature::scan

//======================================================================

bool Feature::Version::consistent() const {
     const char* p = mData.c_str();

     for (; *p && isdigit(*p); p++);

     while (*p) {
          if ( (*p != '.') || /* only char allowed as non-digit */
               (!*++p) || /* can't finish string with it */
               (!isdigit(*p)) ) /* we need at least one digit */
               return false;
          for (; *p && isdigit(*p); p++);
     }
               
     return true;
}

/**
 * Tokenize the string str in a vector of strings
 */
void Feature::Version::tokenize(const string& str,
                                vector<string>& tokens,
                                const string& delimiters)
{
     string::size_type begin, end;
     begin = str.find_first_not_of(delimiters, 0);
     end   = str.find_first_of(delimiters, begin);

     while (begin < str.size() || end < str.size()) {
          tokens.push_back(str.substr(begin, end - begin));
          begin = str.find_first_not_of(delimiters, end);
          end   = str.find_first_of(delimiters, begin);
     }
}

//======================================================================

/**
 * Compare two regular versions to each other
 *
 * XXX: this uses a tokenizer-based algorithm, but could probably be
 * just strcmp(). I can't formally prove that strcmp() works, so I prefer
 * working however suboptimal (eg. 2 passes instead of one) solution.
 */
Feature::Comparison Feature::Version::compare(const Version& toCompare ) const {
     vector<string> thisTokens, otherTokens;
     tokenize(mData, thisTokens, ".");
     tokenize(toCompare, otherTokens, ".");

     vector<string>::iterator t = thisTokens.begin(),
          o = otherTokens.begin();

     int thisVers, otherVers;
     
     for(; t < thisTokens.end() && o < otherTokens.end(); t++, o++) {
          if ( (thisVers = atoi(t->c_str())) < (otherVers = atoi(o->c_str())))
               return cThisOlder;
          else if (thisVers > otherVers)
               return cThisNewer;
     }     
     if (thisTokens.size() < otherTokens.size())
          return cThisOlder;
     else if (thisTokens.size() > otherTokens.size())
          return cThisNewer;
     else
          return cEqual;
}
