/*-
 * Copyright (c) 1998-2002 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/tcl/HSystem.cc,v 1.35 2002/04/26 19:55:26 okumoto Exp $
 */

// $Id$

#include "HSystem.hh"
#include "String.hh"

#include <cerrno>
#include <ctime>
#include <fstream>
#include <stdexcept>

void
HSystem::gen_file(ostream& out) const
{
    gen_header(out);
    gen_classes(out);
    gen_register_command_access(out);
    gen_footer(out);
}

void
HSystem::gen_header(ostream& out) const
{
    out <<
	"/*\n"
	" * This file is automatically generated by tcl_interface_gen, do not edit\n"
	" */\n"
	"\n"
	"#ifndef TclLanguageInterface_hh\n"
	"#include \"tcl/TclLanguageInterface.hh\"\n"
	"#endif\n"
	"\n"
	"#include <tcl.h>\n"
	"#include <iostream>\n"
	"#include <typeinfo>\n"
	"\n"
	"#define H_LIB_TCL_INTERFACE 1\n"
	"\n"
	"\n";

    for (Includes::const_iterator ii = includes.begin(); ii != includes.end(); ++ii) {
	out << "#include \"" << *ii << "\"\n";
    }
    out <<
	"\n"
	"\n";
}

void
HSystem::gen_classes(ostream& out) const
{
    for (Descriptions::const_iterator di = descriptions.begin(); di != descriptions.end(); ++di) {
	const ClassDesc	*cd = *di;

	cout << cd->name << endl;	// debug
	gen_class(out, *cd);
    }
}

void
HSystem::gen_register_command_access(ostream& out) const
{
    out <<
	"void "
	"register_command_access_tcl()\n"
	"{\n";

    for (Descriptions::const_iterator di = descriptions.begin(); di != descriptions.end(); ++di) {
	const ClassDesc	*cd = *di;
	out << "\ttclLanguageInterface->registerCommandAccess( typeid( "
	    << cd->name << " ).name(), new "
	    << replace_chars(cd->name, ':', '_') << "Access() );\n";
    }

    out <<
	"\n";

    for (Descriptions::const_iterator di = descriptions.begin(); di != descriptions.end(); ++di) {
	const ClassDesc	*cd = *di;

	for (const MethodDesc* md = cd->methods; md->name != NULL; ++md) {
	    if (!md->dynamic()) {
		out << "\ttclLanguageInterface->create_command( \""
		    << cd->name << "::"
		    << md->name << "\", &"
		    << cd->name << "Access::staticMethod_"
		    << md->name << ", 0 );\n";
		if (cd->name == md->name)
		    out << "\ttclLanguageInterface->create_command( \""
			<< cd->name << "\", &"
			<< cd->name << "Access::staticMethod_"
			<< md->name << ", 0 );\n";
	    }
	}
    }

    out <<
	"}\n"
	"\n"
	"\n";
}

void
HSystem::gen_footer(ostream& out) const
{
    out <<
	"/* end */\n"
	"\n";
}

void
HSystem::gen_class(ostream& out, const ClassDesc &cd) const
{
    const string	name = replace_chars(cd.name, ':', '_');

    /*
     * Generate class head
     */
    out << "class " << name << "Access : ";
    if (cd.parents[0]) {
	for (const ClassDesc *const* pd = cd.parents; *pd; ++pd) {
	    const char *pname = (*pd)->name;
	    if (pd != cd.parents) out << ", ";
	    out << "public " << replace_chars(pname, ':', '_') << "Access";
	}
    } else {
	out << "virtual public TclLanguageInterface::CommandAccess";
    }

    /*
     * Generate class members
     */
    out <<
	"\n"
	"{\n"
	"  public:\n"
	"\tvirtual Tcl_ObjCmdProc* methodCmd() const\n"
	"\t\t{\n"
	"\t\t\treturn(&" << name << "Access::methodCommand);\n"
	"\t\t}\n"
	"\n"
	"  protected:\n";

    gen_methodCommand(out);
    gen_dynamic_methods(out, cd);

    if (cd.staticMethods() > 0) {
	out << "  public:\n";

	for (const MethodDesc* md = cd.methods; md->name != NULL; ++md) {
	    if (!md->dynamic()) {
		cout << "\t" << md->name << "\n";	// debug

		gen_method(out, cd, *md);
	    }
	}
    }

    if (cd.has_callbacks()) {
	gen_callback_interfaces(out, cd);
    }

    out <<
	"}; /* class " << name << "Access */\n"
	"\n"
	"\n";
}

string
HSystem::gen_result_to_obj(MethodDesc::Value t) const
{
    switch ((MethodDesc::ValueType)t) {
    case MethodDesc::vtVoid:
	return("Tcl_NewObj()");

    case MethodDesc::vtChar:
    case MethodDesc::vtUChar:
    case MethodDesc::vtShort:
    case MethodDesc::vtUShort:
    case MethodDesc::vtInt:
    case MethodDesc::vtUInt:
    case MethodDesc::vtLong:
    case MethodDesc::vtULong:
    case MethodDesc::vtLongLong:
    case MethodDesc::vtULongLong:
    case MethodDesc::vtBool:
    case MethodDesc::vtFloat:
    case MethodDesc::vtDouble:
    case MethodDesc::vtLongDouble:
    case MethodDesc::vtString:
    case MethodDesc::vtObject:
	return("tclLanguageInterface->cxx2tcl( result_object )");

    case MethodDesc::vtList:
    case MethodDesc::vtListRef:
    case MethodDesc::vtListPtr:
	return("tclLanguageInterface->list2tcl( interp, result_object )");

    case MethodDesc::vtLanguageInterface:
	throw LanguageInterface::Error(string("vtLanguageInterface cannot be used as method return type for method ") + (const char*)t);

    case MethodDesc::vtCallbackData:
	return("tclLanguageInterface->callbackData2tcl( interp, result_object )");

    default:
	return("unknown");
    }
}

string
HSystem::gen_method_argument(MethodDesc::Value t,
			     int argno, int method_argno,
			     bool last_arg_in_set_method) const
{
	const string	xargno		= to_string(argno);
	const string	xmethod_argno	= to_string(method_argno);
	string		result;

	if (t.optional() &&
	    t != MethodDesc::vtList &&
	    t != MethodDesc::vtListRef &&
	    t != MethodDesc::vtListPtr &&
	    !last_arg_in_set_method) {
		result += "( objc - 1 < " + xargno + " ) ? " + t.defaultValue() + " : ";
	}

	switch ((MethodDesc::ValueType)t) {
	case MethodDesc::vtVoid:
	    return("void");

	case MethodDesc::vtChar:
	case MethodDesc::vtUChar:
	    result += "*Tcl_GetStringFromObj( objv[" + xargno + "], 0 )";
	    break;

	case MethodDesc::vtShort:
	case MethodDesc::vtUShort:
	case MethodDesc::vtInt:
	case MethodDesc::vtUInt:
	    result += "TclLanguageInterface::tcl_obj2int";
	    result += "( interp, objv[" + xargno + "], method, " + xmethod_argno + " )";
	    break;

	case MethodDesc::vtLong:
	case MethodDesc::vtULong:
	case MethodDesc::vtLongLong:
	case MethodDesc::vtULongLong:
	    result += "TclLanguageInterface::tcl_obj2long";
	    result += "( interp, objv[" + xargno + "], method, " + xmethod_argno + " )";
	    break;

	case MethodDesc::vtBool:
	    result += "TclLanguageInterface::tcl_obj2bool";
	    result += "( interp, objv[" + xargno + "], method, " + xmethod_argno + " )";
	    break;

	case MethodDesc::vtFloat:
	case MethodDesc::vtDouble:
	case MethodDesc::vtLongDouble:
	    result += "TclLanguageInterface::tcl_obj2double";
	    result += "( interp, objv[" + xargno + "], method, " + xmethod_argno + " )";
	    break;

	case MethodDesc::vtString:
	    result += "TclLanguageInterface::to_string( objv[" + xargno + "] )";
	    break;

	case MethodDesc::vtObject:
	    result += "tclLanguageInterface->tcl2object";
	    result += "( objv[" + xargno + "], (" + (const char*)t + "*)0, method, " + xmethod_argno + " )";
	    break;

	case MethodDesc::vtList:
	case MethodDesc::vtListRef:
	case MethodDesc::vtListPtr:
	    if (t.optional() && !last_arg_in_set_method) {
		result += "( objc - 1 < " + xargno + " ) ? ( a" + to_string(method_argno - 1) + " = " + t.defaultValue() + " ) : ";
	    }
	    result += "tclLanguageInterface->tcl2list";
	    result += "( a" + to_string(method_argno - 1) + ", interp, objv[" + xargno + "], method, " + xmethod_argno + " )";
	    break;

	case MethodDesc::vtLanguageInterface:
	    return("tclLanguageInterface");

	case MethodDesc::vtCallbackData:
	    result += "tclLanguageInterface->tcl2callbackData";
	    result += "( interp, objv[" + xargno + "] )";
	    break;

	default:
	    return("unknown");
	}

	return(result);
}

void
HSystem::gen_methodCommand(ostream& out) const
{
    out <<
	"\tstatic int "
	"methodCommand( ClientData obj, Tcl_Interp *interp, "
		        "int objc, Tcl_Obj *const objv[] )\n"
	"\t\t{\n"
	"\t\t\ttry {\n"
	"\t\t\t	Pointer<LanguageInterface::Object> object = static_cast<LanguageInterface::Object*>( obj );\n"
	"\t\t\t	if ( !object )\n"
	"\t\t\t		return error( interp, \"Internal error in methodCommand: Invalid object \", TclLanguageInterface::to_string( objv[0] ), 0 );\n"
	"\t\t\t	const char* interpreter_name = TclLanguageInterface::to_string( objv[1] );\n"
	"\t\t\t	const bool master_interpreter = !strncmp( interpreter_name, \"master\", 6 );\n"
	"\t\t\t	/* const bool safe_interpreter = !strncmp( interpreter_name, \"safe\", 4 ); */\n"
	"\t\t\t	const char* method = TclLanguageInterface::to_string( objv[2] );\n"
	"\t\t\t	return methodCommand_object( object, method, master_interpreter, interp, objc, objv );\n"
	"\t\t\t}\n"
	"\t\t\tcatch ( exception& err ) {\n"
	"\t\t\t	return error( interp, err.what(), 0 );\n"
	"\t\t\t}\n"
	"\t\t}\n"
	"\n";
}

void
HSystem::gen_callback_interfaces(ostream& out, const ClassDesc &cd) const
{
    /*
     * Generate member function declaration
     */
    out <<
	"\tvirtual void "
	"activate_callback( "
			"Tcl_Interp* interp, Tcl_Obj* aCommand, "
			"Tcl_Obj* aUserData, Tcl_Obj* aArg, "
			"LanguageInterface::SpecificCallbackData& aSpecificCallbackData )\n";

    /*
     * Generate member function body
     */
    out <<
	"\t\t{\n"
	"\t\t\t/* cerr << \"" << cd.name << " activate_callback\" << endl; */\n"
	"\t\t\t";

    for (const CallBackDesc *cbd = cd.callbacks; cbd->type; ++cbd) {
	out <<
	    "if ( dynamic_cast<" << cbd->type << "*>( &aSpecificCallbackData ) != 0 ) {\n"
	    "\t\t\t\t" << cbd->type << "& scd = dynamic_cast<" << cbd->type << "&>( aSpecificCallbackData );\n"
	    "\t\t\t\tTcl_Obj* toEval = Tcl_NewObj();\n"
	    "\t\t\t\tif ( Tcl_ListObjAppendElement( interp, toEval, aCommand ) != TCL_OK )\n"
	    "\t\t\t\t\tthrow LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n";

	for (const char* const* arg = cbd->args; *arg; ++arg) {
	    out <<
		"\t\t\t\tif ( Tcl_ListObjAppendElement( interp, toEval, tclLanguageInterface->cxx2tcl( scd." << *arg << " ) ) != TCL_OK )\n"
		"\t\t\t\t\tthrow LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n";
	}
	out <<
	    "\t\t\t\tif ( aArg && Tcl_ListObjAppendElement( interp, toEval, aArg ) != TCL_OK )\n"
	    "\t\t\t\t\tthrow LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n"
	    "\t\t\t\tif ( aUserData && Tcl_ListObjAppendElement( interp, toEval, aUserData ) != TCL_OK )\n"
	    "\t\t\t\t\tthrow LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n"
	    "\t\t\t\tTcl_IncrRefCount( toEval );\n"
	    "\t\t\t\tif ( Tcl_EvalObj( interp, toEval ) != TCL_OK )\n"
	    "\t\t\t\t\tthrow LanguageInterface::Error( string( \"Error evaluating Tcl command: \" ) + Tcl_GetStringResult( interp ) );\n"
	    "\t\t\t\tTcl_DecrRefCount( toEval );\n"
	    "\t\t\t\ttclLanguageInterface->tcl2cxx( interp, Tcl_GetObjResult( interp ), scd.result() );\n"
	    "\t\t\t}\n"
	    "\t\t\telse ";
    }

    out <<
	"{\n"
	"\t\t\t\tthrow LanguageInterface::Error( string( \"Callback invocation for \" ) + typeid( aSpecificCallbackData ).name() + \" is not supported\" );\n"
	"\t\t\t}\n"
	"\t\t}\n"
	"\n";
}

void
HSystem::gen_dynamic_methods(ostream& out, const ClassDesc &cd) const
{
    const string	name = cd.name;
    const string	cxx_name = cd.cxx_name;

    /*
     * Generate member function declaration
     */
    out <<
	"\tstatic int "
	"methodCommand_object( LanguageInterface::Object* object, "
			      "const char* method, bool master_interpreter, "
			      "Tcl_Interp *interp, int objc, "
			      "Tcl_Obj *const objv[] )\n";

    /*
     * Generate member function body
     */
    out <<
	"\t\t{\n"
	"\t\t\t" << cxx_name << "* object_t = dynamic_cast<" << cxx_name << "*>( object );\n"
	"\t\t\tif ( object_t == 0 )\n"
	"\t\t\t\treturn error( interp, \"Internal error in methodCommand: Invalid object type (" << name << "expected): \", object_t->objecttypename(), 0 );\n";

    /*
     * Generate command decode code
     */
    out <<
	"\t\t\tif ( !strcmp( method, \"help\" ) ) {\n"
	"\t\t\t\tTcl_SetObjResult( interp, Tcl_NewStringObj( const_cast<char*>( object_t->class_documentation().c_str() ), -1 ) );\n"
	"\t\t\t}\n"
	"\t\t\telse ";

    for (const MethodDesc* method = cd.methods; method->name; ++method) {
	if (method->dynamic()) {
	    cout << "\t" << method->name << "\n";	// debug

	    out << "if ( !strcmp( method, \"" << method->name << "\" ) ) {\n";
	    if (method->safeExecution()) {
		out << "\t\t\t\tif ( !master_interpreter ) \n"
		<< "\t\t\t\t\treturn error( interp, \"" << cd.name << "::" << method->name << " cannot be called in slave interpreter\", 0 );\n";
	    }

	    int min_args, max_args;
	    method->numberOfArgs( min_args, max_args );
	    out <<
		"\t\t\t\tif ( objc - 3 < " << min_args << " || objc - 3 > " << max_args << " ) \n"
		"\t\t\t\t\treturn error( interp, \"Wrong number of arguments: \", TclLanguageInterface::to_string( objv[0] ), \" \", method, \"\\n\", object_t->method_documentation( object_t->classDescription(), method ).c_str(), 0 );\n";

	    for (int argno = 0; argno < max_args; ++argno) {
		if (!method->getOrSet() || argno != max_args - 1) {
		    out <<
			"\t\t\t\t" << arg_type(*method, argno);
		    if (arg_islist(*method, argno)) {
			out <<
			    ";\n"
			    "\t\t\t\t" << gen_method_argument(method->args[argno], argno + 3, argno + 1, false) << ";\n";
		    } else {
			out << " = " << gen_method_argument(method->args[argno], argno + 3, argno + 1, false) << ";\n";
		    }
		}
	    }

	    const bool result_present = method->result != MethodDesc::vtVoid;
	    if (method->getOrSet()) {
		if (method->setReturnsOld()) {
		    out << "\t\t\t\tconst " << LanguageInterface::method_result_type( method->result, false ) << " result_object = object_t->" << paramlist(*method, max_args - 1) << ";\n";

		    out <<
			"\t\t\t\tif ( objc - 3 == " << max_args << " ) { /* set requested */\n"
			"\t\t\t\t\t" << arg_type(*method, max_args - 1);
		    if (arg_islist(*method, max_args - 1)) {
			out <<
			    ";\n"
			    "\t\t\t\t\t" << gen_method_argument(method->args[max_args - 1], max_args + 2, max_args, true) << ";\n";
		    } else {
			out << " = " << gen_method_argument(method->args[max_args - 1], max_args + 2, max_args, true) << ";\n";
		    }
		    out <<
			"\t\t\t\t\tobject_t->" << paramlist(*method, max_args) << ";\n"
			"\t\t\t\t}\n";
		} else if (method->setReturnsNew()) {
		    out <<
			"\t\t\t\tif ( objc - 3 == " << max_args << " ) { /* set requested */\n"
			"\t\t\t\t\t" << arg_type(*method, max_args - 1);
		    if (arg_islist(*method, max_args - 1)) {
			out <<
			    ";\n"
			    "\t\t\t\t\t" << gen_method_argument(method->args[max_args - 1], max_args + 2, max_args, true) << ";\n";
		    } else {
			out << " = " << gen_method_argument(method->args[max_args - 1], max_args + 2, max_args, true) << ";\n";
		    }
		    out <<
			"\t\t\t\t\tobject_t->" << paramlist(*method, max_args) << ";\n"
			"\t\t\t\t}\n"
			"\t\t\t\tconst " << LanguageInterface::method_result_type( method->result, false ) << " result_object = object_t->" << paramlist(*method, max_args - 1) << ";\n";
		} else {
		    out <<
			"\t\t\t\tif ( objc - 3 == " << max_args << " ) { /* set requested */\n"
			"\t\t\t\t\t" << arg_type(*method, max_args - 1);
		    if (arg_islist(*method, max_args - 1)) {
			out <<
			    ";\n"
			    "\t\t\t\t\t" << gen_method_argument(method->args[max_args - 1], max_args + 2, max_args, true) << ";\n";
		    } else {
			out <<
			    " = " << gen_method_argument(method->args[max_args - 1], max_args + 2, max_args, true) << ";\n";
		    }
		    out << "\t\t\t\t\tobject_t->" << paramlist(*method, max_args) << ";\n"
			"\t\t\t\t\tTcl_ResetResult( interp );\n"
			"\t\t\t\t}\n"
			"\t\t\t\telse { /* get requested */\n"
			"\t\t\t\t\tconst " << LanguageInterface::method_result_type( method->result, false ) << " result_object = object_t->" << paramlist(*method, max_args - 1) << ";\n"
			"\t\t\t\t\tTcl_SetObjResult( interp, " << gen_result_to_obj( method->result ) << " );\n"
			"\t\t\t\t}\n";
		}
	    } else {
		out << "\t\t\t\t";
		if (result_present)
			out << "const " << LanguageInterface::method_result_type( method->result, false ) << " result_object = ";
		out << "object_t->" << paramlist(*method, max_args) << ";\n";
	    }

	    if (result_present) {
		if (!method->getOrSet() || method->setReturnsNew() || method->setReturnsOld())
		    out <<
			"\t\t\t\tTcl_SetObjResult( interp, " << gen_result_to_obj( method->result ) << " );\n";
	    } else {
		out << "\t\t\t\tTcl_ResetResult( interp );\n";
	    }
	    out <<
		"\t\t\t}\n"
		"\t\t\telse ";
	}
    }

    if (cd.parents[0]) {
	out << "if ( ";
	for (const ClassDesc *const* pd = cd.parents; *pd; ++pd) {
		if (pd != cd.parents) out << " && ";
		out << replace_chars( (*pd)->name, ':', '_' ) << "Access::methodCommand_object( object, method, master_interpreter, interp, objc, objv ) == TCL_ERROR";
	}
	out <<
	    " ) {\n"
	    "\t\t\t\t return(TCL_ERROR);\n"
	    "\t\t\t}\n";
    } else {
	out <<
	    "{\n"
	    "\t\t\t\t /* cerr << \"methodCommand: \" << (void*)objv[0] << \" \" << TclLanguageInterface::to_string( objv[0] ) << ' ' << method << endl; */\n"
	    "\t\t\t\t return error( interp, \"Unknown method \", method, \" for object of type \", object_t->objecttypename(), 0 );\n"
	    "\t\t\t}\n";
    }
    out <<
	"\t\t\treturn(TCL_OK);\n"
	"\t\t}\n"
	"\n";
}

void
HSystem::gen_method(ostream& out, const ClassDesc &cd, const MethodDesc &md) const
{
    const string	cxx_name = cd.cxx_name;
    const string	mname = string(cd.name) + "::" + md.name;

    out <<
	"\tstatic int "
	"staticMethod_" << md.name << "( ClientData /*obj*/, "
				"Tcl_Interp *interp, int objc, "
				"Tcl_Obj *const objv[] )\n"
	"\t\t{\n"
	"\t\t\ttry {\n";

    if (md.safeExecution()) {
	out <<
	    "\t\t\t\tconst char* interpreter_name = TclLanguageInterface::to_string( objv[1] );\n"
	    "\t\t\t\tconst bool master_interpreter = !strncmp( interpreter_name, \"master\", 6 );\n"
	    "\t\t\t\t/* const bool safe_interpreter = !strncmp( interpreter_name, \"safe\", 4 ); */\n"
	    "\t\t\t\tif ( !master_interpreter ) \n"
	    "\t\t\t\t\treturn error( interp, \"" << mname << " cannot be called in slave interpreter\", 0 );\n";
    }

    int min_args;
    int max_args;
    md.numberOfArgs(min_args, max_args);
    int max_args_in_call = (md.args[max_args-1] == MethodDesc::vtLanguageInterface) ? max_args - 1 : max_args;
    out <<
	"\t\t\t\tif ( objc - 2 < " << min_args << " || objc - 2 > " << max_args_in_call << " ) \n"
	"\t\t\t\t\treturn error( interp, \"Wrong number of arguments: \", \"" << mname << "\", \" (was \", to_string( objc - 2 ).c_str(), \", expected " << ( min_args < max_args_in_call ? to_string( min_args ) + " to " + to_string( max_args_in_call ) : to_string( max_args_in_call ) ) << ")\", 0 );\n"
	"\t\t\t\tconst char* method = \"" << md.name << "\";\n";

    for (int argno = 0; argno < max_args; ++argno) {
	if (arg_islist(md, argno)) {
	    out <<
		"\t\t\t\t" << arg_type(md, argno) << ";\n"
		"\t\t\t\t" << gen_method_argument(md.args[argno], argno + 2, argno + 1, false) << ";\n";
	} else {
	    out <<
		"\t\t\t\t" << arg_type(md, argno) << " = " << gen_method_argument(md.args[argno], argno + 2, argno + 1, false) << ";\n";
	}
    }
    if (md.result == MethodDesc::vtVoid) {
	out <<
	    "\t\t\t\t" << cxx_name << "::" << paramlist(md, max_args) << ";\n"
	    "\t\t\t\tTcl_ResetResult( interp );\n"
	    "\t\t\t\treturn(TCL_OK);\n";
    } else {
	out <<
	    "\t\t\t\t" << LanguageInterface::method_result_type(md.result, false) << " result_object = " << cxx_name << "::" << paramlist(md, max_args) << ";\n"
	    "\t\t\t\tTcl_SetObjResult( interp, " << gen_result_to_obj(md.result) << " );\n"
	    "\t\t\t\treturn(TCL_OK);\n";
    }

    out <<
	"\t\t\t}\n"
	"\t\t\tcatch ( exception& err ) {\n"
	"\t\t\t\treturn error( interp, err.what(), 0 );\n"
	"\t\t\t}\n"
	"\t\t}\n"
	"\n";
}

/*============================================================================
 * This is the parser code which I will link in later.
 *===========================================================================*/
#if 0
#include <fstream>
#include <map>
#include <set>
#include <string>
#include <vector>

extern "C" {
    #include <unistd.h>		/* getopt() */
    #include <string.h>		/* strlen() */
}

class ClassDescParser {
    typedef string			FilePath;
    typedef string			DirPath;
    typedef string			SysName;
    typedef string			ClassDesc;

    typedef set<FilePath>		FileSet;
    typedef vector<ClassDesc>		ClassDescV;
    typedef map<FilePath, ClassDescV>	MapV;
    typedef map<DirPath, SysName>	Map;

    FileSet	files;
    MapV	classes;
    Map		systems;

public:
    /**
     * Process the contents of the ClassDescription file (*.cd.cc).
     *
     * @param filename	name of file to process.
     */
    void
    process_file(const string &filename)
    {
	ifstream	file(filename.c_str());

	char		buff[4096];
	while (file.getline(buff, sizeof(buff))) {
	    string	line(buff);

	    /* remove C++ style comments from line */
	    string::size_type pos = line.find("//");
	    if (pos != string::npos) {
		line.erase(pos);
	    }


	   /*
	    * LanguageInterface::Object::ClassDescription ([^\s]+)::sClassDescription[^A-Za-z0-9]/
	    */
	    string prefix("LanguageInterface::Object::ClassDescription ");
	    string suffix("::sClassDescription");

	    /* Find lines which contain prefix */
	    pos = line.find(prefix);
	    if (pos == string::npos) {
		continue;	/* line does not contain prefix */
	    }
	    string	cd = line.substr(pos + prefix.length());

	    pos = cd.find(suffix);

	    /* class description must also contain the suffix */
	    assert(pos != string::npos);

	    /* char following suffix should not be alpha or number */
	    assert(!isalnum(cd[pos + suffix.length()]));

	    cd.erase(pos);

	    assert(cd.find_first_of(" \t") == string::npos);

	    classes[filename].push_back(cd);
	}

	file.close();

	files.insert(filename);
    }

    /**
     * Output the build_systems_*.cc file.
     *
     * @param prog		Name of generator program.
     * @param the_only_system	(Don't know yet.)
     */
    void
    gen_file(const string &prog, const string *the_only_system)
    {
	cout << "/*\n"
		" * This file is automatically generated by " << prog << ",\n"
		" * do not edit!\n"
		" */\n"
		"\n"
		"#ifndef HSystem_hh\n"
		"#include \"HSystem.hh\"\n"
		"#endif\n"
		"\n"
		"#define H_BUILD_SYSTEMS 1\n"
		"\n";

	for (MapV::iterator i = classes.begin(); i != classes.end(); ++i) {
	    FilePath		filename = i->first;
	    string::size_type	pos = filename.rfind('/');

	    assert(pos != string::npos); /* contains at least 1 slash */
	    string		system = filename.substr(0, pos);
	    string		file   = filename.substr(pos + 1);
	
	    if (the_only_system == NULL) {
		pos = system.rfind('/');
		assert(pos != string::npos);	/* contains 1 slash */

		systems[system] = system.substr(pos + 1);
	    } else {
		systems[*the_only_system] = *the_only_system;
	    }

	    pos = system.rfind('/');
	    pos = system.rfind('/', pos);
	    string	systemdir = system.substr(pos + 1);

	    /* remove ".cd.cc" suffix */
	    pos = file.find(".cd.cc");
	    assert(pos != string::npos); /* contains ".cd.cc" */
	    assert(file.length() == pos + strlen(".cd.cc"));	/* at end */
	    file.erase(pos);

	    if (systemdir == "common") {
		cout << "#ifndef " << file << "_hh\n"
			"#include \"" << file << ".hh\"\n"
			"#endif\n"
			"\n";
	    } else {
		cout << "#ifndef " << file << "_hh\n"
			"#include \"" << systemdir << "/" << file << ".hh\"\n"
			"#endif\n"
			"\n";
	    }
	}

	cout << "void\n"
		"build_systems(Systems& systems)\n"
		"{\n";

	for (Map::iterator i = systems.begin(); i != systems.end(); ++i) {
	    DirPath	system	= i->first;
	    SysName	sysname	= i->second;

	    cout <<
		"\tpair<Systems::iterator,bool> s = systems.insert(HSystem(\"" << sysname << "\"));\n"
		"\tif (!s.second)\n"
		"\t	throw runtime_error(string(\"System \") + s.first->name + \" already exists\");\n"
		"\tHSystem& system = const_cast<HSystem&>(*s.first);\n";

	    for (FileSet::iterator j = files.begin(); j != files.end(); ++j) {
		FilePath filename = *j;

		if ((the_only_system != NULL) || (filename.substr(0, system.length() + 1) == (system + '/'))) {
		    ClassDescV cl = classes[filename];
		    for (ClassDescV::iterator k = cl.begin(); k != cl.end(); ++k) {
			cout <<
			    "\tsystem.add_class(" << *k << "::thisClassDescription(),\n"
			    "\t\t\t" << *k << "::fileToInclude());\n";
		    }
		}
	    }
	}

	cout << "} /* build systems */ \n"
		"\n";
    }
};

/**
 * Outputs usage information.
 *
 * @param prog		Name of binary
 */
static void
usage(const string &prog)
{
    cout << "Usage: " << prog << " -s <system> <file-to-scan> ...\n";
}

/**
 * The main program.
 *
 * It parses the command line arguments, and checks for input files.
 */
int
main(int argc, char *argv[])
{
    /*************************************************************************
     * Parse command line args
     *************************************************************************/
    const string	prog(argv[0]);

    string		*the_only_system = NULL;

    int c;
    while ((c = getopt(argc, argv, "s:")) != -1) {
	switch (c) {
	case 's':
	    the_only_system = new string(optarg);
	    break;
	case '?':
	default:
	    usage(prog);
	    return(1);	/* bad cli options */
	}  
    }
    argc -= optind;
    argv += optind;

    if (argc < 1) {
	usage(prog);
	return(1);	/* missing input files */
    }

    /*************************************************************************
     * process files
     *************************************************************************/

    ClassDescParser	parser;

    for (int i = 0; i < argc; ++i) {
	parser.process_file(argv[i]);
    }

    parser.gen_file(prog, the_only_system);

    delete the_only_system;
    return(0);
}

#endif
