/*
 * $Id: Module.cc,v 1.4 2002/07/13 10:24:41 okumoto Exp $
 */
#include "Exceptions.hh"
#include "Library.hh"
#include "Module.hh"
#include "Program.hh"

#include <fstream>
#include <string>
#include <vector>

/**
 * Constructor.
 *
 * @param src	Directory containing source files.
 * @param obj	Destination of generated output files.
 * @param dir	Directory being processed.
 */
Module::Module(const string &src, const string &obj, const string &dir)
    : srcdir(src), objdir(obj), dirname(dir), oc(NULL)
{
    /*
     * open module description file
     */
    filename = dirname.empty() ? "module.m4" : dirname + "/module.m4";

    ifstream	input(filename.c_str());
    if (!input) {
	throw runtime_error("Error opening " + filename);
    }

    /*
     * parse module description file line by line
     */
    int		linenum = 0;
    char	buff[4096];
    while (input.getline(buff, 4096)) {
	const string	line(buff);

	if (line.empty()) {
	    ++linenum;
	    continue;	/* ignore blank lines */
	}
	if (line[0] == '#') {
	    ++linenum;
	    continue;	/* ignore comments */
	}

	parse_line(line, linenum);
	++linenum;
    }
}

/**
 * Extract the action token and the argument from the input line.
 *
 * @param line	Line to parse.
 * @param linenum	Line number of parsed line.
 */
void
Module::parse_line(const string &line, int linenum)
{
    struct Keyword {
	const char	*str;
	int		len;
	ModuleToken	token;
    };

    static const Keyword	keywords[] = {
	{ "cfile",	5, tCFILE },
	{ "cflags",	6, tCFLAGS },
	{ "ldflags",	7, tLDFLAGS },
	{ "library",	7, tLIBRARY },
	{ "program",	7, tPROGRAM },
	{ "subdir",	6, tSUBDIR },
	{ "swigfile",	8, tSWIGFILE },
	{ "uselib",	6, tUSELIB },
	{ NULL,		0, tNULL }
    };

    /*
     * Find keyword from table
     */
    const Keyword	*k;
    for (k = keywords; k->str != NULL; ++k) {
	if (line.substr(0, k->len) == k->str) {
	    break;
	}
    }
    if (k->str == NULL) {
	throw SyntaxError("missing keyword", linenum, filename);
    }

    /* arg is everything between the parentheses */
    string::size_type	pos;
    if (line[k->len] != '(') {
	throw SyntaxError("missing '('", linenum, filename);
    }
    if ((pos = line.rfind(')')) == string::npos) {
	throw SyntaxError("missing ')'", linenum, filename);
    }

    decode_action(k->token,
		  line.substr(k->len + 1, pos - k->len - 1),
		  linenum);
}

/**
 * Perform the action associated with the token.
 *
 * @param token	Module token extracted from input line.
 * @param arg	Argument extracted from input line.
 * @param linenum	Line number of parsed line.
 */
void
Module::decode_action(ModuleToken token, const string &arg, int linenum)
{
    if (oc == NULL) {
	switch (token) {
	case tLIBRARY: oc = new Library(srcdir, objdir, dirname, arg); break;
	case tPROGRAM: oc = new Program(srcdir, objdir, dirname, arg); break;
	case tSUBDIR:
		if (dirname.empty()) {
		    sd.push_back(arg);
		} else {
		    sd.push_back(dirname + '/' + arg);
		}
		break;

	default:
	    throw Error("Module type not set", linenum, filename);
	    break;
	}
    } else {
	switch (token) {
	case tCFILE:
	    {
		string::size_type	pos;

		if ((pos = arg.rfind('.')) == string::npos) {
		    throw SyntaxError("missing '.'", linenum, filename);
		}

		string base = dirname + '/' + arg.substr(0, pos);
		string suffix = arg.substr(pos + 1);

		if (suffix == "c") {
		    oc->add_cfile(base);
		} else if (suffix == "cc") {
		    oc->add_ccfile(base);
		} else {
		    throw Error("incorrect file suffix", linenum, filename);
		}
	    }
	    break;

	case tCFLAGS: 	oc->append_cflags(arg); break;

	case tLDFLAGS:	oc->append_ldflags(arg); break;

	case tSUBDIR:
		if (dirname.empty()) {
		    sd.push_back(arg);
		} else {
		    sd.push_back(dirname + '/' + arg);
		}
		break;

	case tSWIGFILE:
	    {
		string::size_type	pos;

		if ((pos = arg.rfind('.')) == string::npos) {
		    throw SyntaxError("missing '.'", linenum, filename);
		}

		string base = dirname + '/' + arg.substr(0, pos);
		string suffix = arg.substr(pos + 1);

		if (suffix == "ii") {
		    oc->add_swigfile(base);
		} else {
		    throw Error("incorrect file suffix", linenum, filename);
		}
	    }
	    break;

	case tUSELIB:	oc->uselib(arg); break;

	default:	assert(0); break;
	}
    }
}

